@@ -69,31 +69,6 @@ msDrawRasterLayerGDAL_16BitClassification(
   int src_xoff, int src_yoff, int src_xsize, int src_ysize,
   int dst_xoff, int dst_yoff, int dst_xsize, int dst_ysize );
 
-#ifdef USE_GD
-static void Dither24to8( GByte *pabyRed, GByte *pabyGreen, GByte *pabyBlue,
-                         GByte *pabyDithered, int xsize, int ysize,
-                         int bTransparent, colorObj transparentColor,
-                         gdImagePtr gdImg );
-/*
- * Stuff for allocating color cubes, and mapping between RGB values and
- * the corresponding color cube index.
- */
-
-static int allocColorCube(mapObj *map, gdImagePtr img, int *panColorCube);
-#endif
-
-
-#define RED_LEVELS 5
-#define RED_DIV 52
-/* #define GREEN_LEVELS 7 */
-#define GREEN_LEVELS 5
-/* #define GREEN_DIV 37 */
-#define GREEN_DIV 52
-#define BLUE_LEVELS 5
-#define BLUE_DIV 52
-
-#define RGB_LEVEL_INDEX(r,g,b) ((r)*GREEN_LEVELS*BLUE_LEVELS + (g)*BLUE_LEVELS+(b))
-#define RGB_INDEX(r,g,b) RGB_LEVEL_INDEX(((r)/RED_DIV),((g)/GREEN_DIV),((b)/BLUE_DIV))
 
 /*
  * rasterBufferObj setting macros.
@@ -131,12 +106,6 @@ int msDrawRasterLayerGDAL(mapObj *map, layerObj *layer, imageObj *image,
   int bHaveRGBNoData = FALSE;
   int nNoData1=-1,nNoData2=-1,nNoData3=-1;
   rasterBufferObj *mask_rb = NULL;
-#ifdef USE_GD
-  int   anColorCube[256];
-  int cmt=0;
-  /*make sure we don't have a truecolor gd image*/
-  assert(!rb || rb->type != MS_BUFFER_GD || !gdImageTrueColor(rb->data.gd_img));
-#endif
   if(layer->mask) {
     int ret;
     layerObj *maskLayer = GET_LAYER(map, msGetLayerIndex(map,layer->mask));
@@ -248,21 +217,21 @@ int msDrawRasterLayerGDAL(mapObj *map, layerObj *layer, imageObj *image,
     urx = GEO_TRANS(adfInvGeoTransform+0,copyRect.maxx,copyRect.maxy);
     ury = GEO_TRANS(adfInvGeoTransform+3,copyRect.maxx,copyRect.maxy);
 
-    src_xoff = MAX(0,(int) floor(llx+0.5));
-    src_yoff = MAX(0,(int) floor(ury+0.5));
-    src_xsize = MIN(MAX(0,(int) (urx - llx + 0.5)),
+    src_xoff = MS_MAX(0,(int) floor(llx+0.5));
+    src_yoff = MS_MAX(0,(int) floor(ury+0.5));
+    src_xsize = MS_MIN(MS_MAX(0,(int) (urx - llx + 0.5)),
                     GDALGetRasterXSize(hDS) - src_xoff);
-    src_ysize = MIN(MAX(0,(int) (lly - ury + 0.5)),
+    src_ysize = MS_MIN(MS_MAX(0,(int) (lly - ury + 0.5)),
                     GDALGetRasterYSize(hDS) - src_yoff);
 
     /* We want very small windows to use at least one source pixel (#4172) */
     if( src_xsize == 0 && (urx - llx) > 0.0 ) {
       src_xsize = 1;
-      src_xoff = MIN(src_xoff,GDALGetRasterXSize(hDS)-1);
+      src_xoff = MS_MIN(src_xoff,GDALGetRasterXSize(hDS)-1);
     }
     if( src_ysize == 0 && (lly - ury) > 0.0 ) {
       src_ysize = 1;
-      src_yoff = MIN(src_yoff,GDALGetRasterYSize(hDS)-1);
+      src_yoff = MS_MIN(src_yoff,GDALGetRasterYSize(hDS)-1);
     }
 
     if( src_xsize == 0 || src_ysize == 0 ) {
@@ -282,11 +251,11 @@ int msDrawRasterLayerGDAL(mapObj *map, layerObj *layer, imageObj *image,
 
     dst_lrx = (int) ((copyRect.maxx - mapRect.minx) / map->cellsize + 0.5);
     dst_lry = (int) ((mapRect.maxy - copyRect.miny) / map->cellsize + 0.5);
-    dst_lrx = MAX(0,MIN(image->width,dst_lrx));
-    dst_lry = MAX(0,MIN(image->height,dst_lry));
+    dst_lrx = MS_MAX(0,MS_MIN(image->width,dst_lrx));
+    dst_lry = MS_MAX(0,MS_MIN(image->height,dst_lry));
 
-    dst_xsize = MAX(0,MIN(image->width,dst_lrx - dst_xoff));
-    dst_ysize = MAX(0,MIN(image->height,dst_lry - dst_yoff));
+    dst_xsize = MS_MAX(0,MS_MIN(image->width,dst_lrx - dst_xoff));
+    dst_ysize = MS_MAX(0,MS_MIN(image->height,dst_lry - dst_yoff));
 
     if( dst_xsize == 0 || dst_ysize == 0 ) {
       if( layer->debug )
@@ -321,8 +290,8 @@ int msDrawRasterLayerGDAL(mapObj *map, layerObj *layer, imageObj *image,
   else {
     dst_xoff = src_xoff = 0;
     dst_yoff = src_yoff = 0;
-    dst_xsize = src_xsize = MIN(image->width,src_xsize);
-    dst_ysize = src_ysize = MIN(image->height,src_ysize);
+    dst_xsize = src_xsize = MS_MIN(image->width,src_xsize);
+    dst_ysize = src_ysize = MS_MIN(image->height,src_ysize);
   }
 
   /*
@@ -459,22 +428,6 @@ int msDrawRasterLayerGDAL(mapObj *map, layerObj *layer, imageObj *image,
   if( alpha_band != 0 )
     hBandAlpha = GDALGetRasterBand( hDS, alpha_band );
 
-#ifdef USE_GD
-  /*
-   * Wipe pen indicators for all our layer class colors if they exist.
-   * Sometimes temporary gdImg'es are used in which case previously allocated
-   * pens won't generally apply.  See Bug 504.
-   */
-  if( rb->type == MS_BUFFER_GD ) {
-    int iClass;
-    int iStyle;
-    for( iClass = 0; iClass < layer->numclasses; iClass++ ) {
-      for (iStyle=0; iStyle<layer->class[iClass]->numstyles; iStyle++)
-        layer->class[iClass]->styles[iStyle]->color.pen = MS_PEN_UNSET;
-    }
-  }
-#endif
-
   /*
    * The logic for a classification rendering of non-8bit raster bands
    * is sufficiently different than the normal mechanism of loading
@@ -508,30 +461,17 @@ int msDrawRasterLayerGDAL(mapObj *map, layerObj *layer, imageObj *image,
         pixel.red = i;
         pixel.green = i;
         pixel.blue = i;
-#ifdef USE_GD
-        pixel.pen = i;
-#endif
 
         if(MS_COMPARE_COLORS(pixel, layer->offsite)) {
           sEntry.c1 = 0;
           sEntry.c2 = 0;
           sEntry.c3 = 0;
           sEntry.c4 = 0; /* alpha set to zero */
-        } else if( rb->type != MS_BUFFER_GD ) {
+        } else {
           sEntry.c1 = i;
           sEntry.c2 = i;
           sEntry.c3 = i;
           sEntry.c4 = 255;
-        } else {
-          /*
-          ** This special calculation is intended to use only 128
-          ** unique colors for greyscale in non-truecolor mode.
-          */
-
-          sEntry.c1 = i - i%2;
-          sEntry.c2 = i - i%2;
-          sEntry.c3 = i - i%2;
-          sEntry.c4 = 255;
         }
 
         GDALSetColorEntry( hColorMap, i, &sEntry );
@@ -568,6 +508,8 @@ int msDrawRasterLayerGDAL(mapObj *map, layerObj *layer, imageObj *image,
    */
   if( classified ) {
     int c, color_count;
+    const char* pszRangeColorspace = msLayerGetProcessingKey( layer, "RANGE_COLORSPACE" );
+    colorspace iRangeColorspace;
 
 #ifndef NDEBUG
     cmap_set = TRUE;
@@ -579,8 +521,19 @@ int msDrawRasterLayerGDAL(mapObj *map, layerObj *layer, imageObj *image,
                  "drawGDAL()");
       return -1;
     }
+    
+    if(!pszRangeColorspace || !strcasecmp(pszRangeColorspace, "RGB")) {
+      iRangeColorspace = MS_COLORSPACE_RGB;
+    } else if(!strcasecmp(pszRangeColorspace, "HSL")) {
+      iRangeColorspace = MS_COLORSPACE_HSL;
+    } else {
+      msSetError(MS_MISCERR,
+                 "Unknown RANGE_COLORSPACE \"%s\", expecting RGB or HSL",
+                 "drawGDAL()", pszRangeColorspace);
+      return -1;
+    }
 
-    color_count = MIN(256,GDALGetColorEntryCount(hColorMap));
+    color_count = MS_MIN(256,GDALGetColorEntryCount(hColorMap));
     for(i=0; i < color_count; i++) {
       colorObj pixel;
       int colormap_index;
@@ -596,92 +549,56 @@ int msDrawRasterLayerGDAL(mapObj *map, layerObj *layer, imageObj *image,
       if(!MS_COMPARE_COLORS(pixel, layer->offsite)) {
         c = msGetClass(layer, &pixel, colormap_index);
 
-        if(c == -1) { /* doesn't belong to any class, so handle like offsite*/
-          if( rb->type == MS_BUFFER_GD )
-            cmap[i] = -1;
-        } else {
+        if(c != -1) { /* belongs to any class */
           int s;
 
           /* change colour based on colour range?  Currently we
              only address the greyscale case properly. */
 
-          for(s=0; s<layer->class[c]->numstyles; s++) {
-            if( MS_VALID_COLOR(layer->class[c]->styles[s]->mincolor)
-                && MS_VALID_COLOR(layer->class[c]->styles[s]->maxcolor) )
-              msValueToRange(layer->class[c]->styles[s],
-                             sEntry.c1 );
+          if( MS_VALID_COLOR(layer->class[c]->styles[0]->mincolor) ) {
+            for(s=0; s<layer->class[c]->numstyles; s++) {
+              if( MS_VALID_COLOR(layer->class[c]->styles[s]->mincolor)
+                 && MS_VALID_COLOR(layer->class[c]->styles[s]->maxcolor)) {
+                if( layer->class[c]->numstyles == 1 || (sEntry.c1 >= layer->class[c]->styles[s]->minvalue
+                                               && sEntry.c1 <= layer->class[c]->styles[s]->maxvalue )) {
+                  msValueToRange(layer->class[c]->styles[s], sEntry.c1, iRangeColorspace);
+                  if(MS_VALID_COLOR(layer->class[c]->styles[s]->color)) {
+                    rb_cmap[0][i] = layer->class[c]->styles[s]->color.red;
+                    rb_cmap[1][i] = layer->class[c]->styles[s]->color.green;
+                    rb_cmap[2][i] = layer->class[c]->styles[s]->color.blue;
+                    rb_cmap[3][i] = (layer->class[c]->styles[s]->color.alpha != 255)?(layer->class[c]->styles[s]->color.alpha):(255*layer->class[c]->styles[0]->opacity / 100);
+                    break;
+                  }
+                }
+              }
+            }
           }
-#ifdef USE_GD
-          if( rb->type == MS_BUFFER_GD ) {
-            RESOLVE_PEN_GD(rb->data.gd_img, layer->class[c]->styles[0]->color);
-            if( MS_TRANSPARENT_COLOR(layer->class[c]->styles[0]->color) )
-              cmap[i] = -1;
-            else if( MS_VALID_COLOR(layer->class[c]->styles[0]->color)) {
-              /* use class color */
-              cmap[i] = layer->class[c]->styles[0]->color.pen;
-            } else /* Use raster color */
-              cmap[i] = msAddColorGD(map, rb->data.gd_img, cmt,
-                                     pixel.red, pixel.green, pixel.blue);
-          } else if( rb->type == MS_BUFFER_BYTE_RGBA )
-#endif
-          {
-            if( MS_TRANSPARENT_COLOR(layer->class[c]->styles[0]->color))
-              /* leave it transparent */;
+          else if( MS_TRANSPARENT_COLOR(layer->class[c]->styles[0]->color))
+            /* leave it transparent */;
 
-            else if( MS_VALID_COLOR(layer->class[c]->styles[0]->color)) {
-              rb_cmap[0][i] = layer->class[c]->styles[0]->color.red;
-              rb_cmap[1][i] = layer->class[c]->styles[0]->color.green;
-              rb_cmap[2][i] = layer->class[c]->styles[0]->color.blue;
-              rb_cmap[3][i] = (255*layer->class[c]->styles[0]->opacity / 100);
-            }
+          else if( MS_VALID_COLOR(layer->class[c]->styles[0]->color)) {
+            rb_cmap[0][i] = layer->class[c]->styles[0]->color.red;
+            rb_cmap[1][i] = layer->class[c]->styles[0]->color.green;
+            rb_cmap[2][i] = layer->class[c]->styles[0]->color.blue;
+            rb_cmap[3][i] = (255*layer->class[c]->styles[0]->opacity / 100);
+          }
 
-            else { /* Use raster color */
-              rb_cmap[0][i] = pixel.red;
-              rb_cmap[1][i] = pixel.green;
-              rb_cmap[2][i] = pixel.blue;
-              rb_cmap[3][i] = 255;
-            }
+          else { /* Use raster color */
+            rb_cmap[0][i] = pixel.red;
+            rb_cmap[1][i] = pixel.green;
+            rb_cmap[2][i] = pixel.blue;
+            rb_cmap[3][i] = 255;
           }
         }
-#ifdef USE_GD
-      } else {
-        if( rb->type == MS_BUFFER_GD )
-          cmap[i] = -1;
-#endif
       }
     }
-#ifdef USE_GD
-  } else if( hColorMap != NULL && rb->type == MS_BUFFER_GD ) {
-    int color_count;
-#ifndef NDEBUG
-    cmap_set = TRUE;
-#endif
-
-    color_count = MIN(256,GDALGetColorEntryCount(hColorMap));
-
-    for(i=0; i < color_count; i++) {
-      GDALColorEntry sEntry;
-
-      GDALGetColorEntryAsRGB( hColorMap, i, &sEntry );
-
-      if( sEntry.c4 != 0
-          && (!MS_VALID_COLOR( layer->offsite )
-              || layer->offsite.red != sEntry.c1
-              || layer->offsite.green != sEntry.c2
-              || layer->offsite.blue != sEntry.c3 ) )
-        cmap[i] = msAddColorGD(map, rb->data.gd_img, cmt,
-                               sEntry.c1, sEntry.c2, sEntry.c3);
-      else
-        cmap[i] = -1;
-    }
-#endif
   } else if( hBand2 == NULL && hColorMap != NULL && rb->type == MS_BUFFER_BYTE_RGBA ) {
     int color_count;
 #ifndef NDEBUG
     cmap_set = TRUE;
 #endif
 
-    color_count = MIN(256,GDALGetColorEntryCount(hColorMap));
+    color_count = MS_MIN(256,GDALGetColorEntryCount(hColorMap));
 
     for(i=0; i < color_count; i++) {
       GDALColorEntry sEntry;
@@ -700,12 +617,6 @@ int msDrawRasterLayerGDAL(mapObj *map, layerObj *layer, imageObj *image,
       }
     }
   }
-#ifdef USE_GD
-  else if( rb->type == MS_BUFFER_GD ) {
-    allocColorCube( map, rb->data.gd_img, anColorCube );
-  }
-#endif
-
   /*
    * Allocate imagery buffers.
    */
@@ -803,63 +714,6 @@ int msDrawRasterLayerGDAL(mapObj *map, layerObj *layer, imageObj *image,
     }
   }
 
-#ifdef USE_GD
-  /* -------------------------------------------------------------------- */
-  /*      Single band plus colormap with alpha blending to 8bit.          */
-  /* -------------------------------------------------------------------- */
-  if( hBand2 == NULL && rb->type == MS_BUFFER_GD && hBandAlpha != NULL ) {
-    assert( cmap_set );
-    k = 0;
-
-    for( i = dst_yoff; i < dst_yoff + dst_ysize; i++ ) {
-      int result, alpha;
-
-      for( j = dst_xoff; j < dst_xoff + dst_xsize; j++ ) {
-        alpha = pabyRawAlpha[k];
-
-        result = cmap[pabyRaw1[k++]];
-        if(SKIP_MASK(j,i)) {
-          continue;
-        }
-
-        /*
-        ** We don't do alpha blending in non-truecolor mode, just
-        ** threshold the point on/off at alpha=128.
-        */
-
-        if( result != -1 && alpha >= 128 )
-          rb->data.gd_img->pixels[i][j] = result;
-      }
-    }
-
-    assert( k == dst_xsize * dst_ysize );
-  }
-
-  /* -------------------------------------------------------------------- */
-  /*      Single band plus colormap (no alpha) to 8bit.                   */
-  /* -------------------------------------------------------------------- */
-  else if( hBand2 == NULL  && rb->type == MS_BUFFER_GD ) {
-    assert( cmap_set );
-    k = 0;
-
-    for( i = dst_yoff; i < dst_yoff + dst_ysize; i++ ) {
-      int result;
-
-      for( j = dst_xoff; j < dst_xoff + dst_xsize; j++ ) {
-        result = cmap[pabyRaw1[k++]];
-        if(SKIP_MASK(j,i)) {
-          continue;
-        }
-        if( result != -1 ) {
-          rb->data.gd_img->pixels[i][j] = result;
-        }
-      }
-    }
-
-    assert( k == dst_xsize * dst_ysize );
-  } else
-#endif
-
     /* -------------------------------------------------------------------- */
     /*      Single band plus colormap and alpha to truecolor. (RB)          */
     /* -------------------------------------------------------------------- */
@@ -973,91 +827,6 @@ int msDrawRasterLayerGDAL(mapObj *map, layerObj *layer, imageObj *image,
       }
     }
 
-#ifdef USE_GD
-  /* -------------------------------------------------------------------- */
-  /*      Input is 3 band RGB.  Alpha blending is mixed into the loop     */
-  /*      since this case is less commonly used and has lots of other     */
-  /*      overhead. (GD)                                                  */
-  /* -------------------------------------------------------------------- */
-    else if( hBand3 != NULL && rb->type == MS_BUFFER_GD ) {
-      /* Dithered 24bit to 8bit conversion */
-      if( CSLFetchBoolean( layer->processing, "DITHER", FALSE ) ) {
-        unsigned char *pabyDithered;
-
-        pabyDithered = (unsigned char *) malloc(dst_xsize * dst_ysize);
-        if( pabyDithered == NULL ) {
-          msSetError(MS_MEMERR, "Allocating work image of size %dx%d failed.",
-                     "msDrawRasterLayerGDAL()", dst_xsize, dst_ysize );
-          return -1;
-        }
-
-        Dither24to8( pabyRaw1, pabyRaw2, pabyRaw3, pabyDithered,
-                     dst_xsize, dst_ysize, image->format->transparent,
-                     map->imagecolor, rb->data.gd_img );
-
-        k = 0;
-        for( i = dst_yoff; i < dst_yoff + dst_ysize; i++ ) {
-          for( j = dst_xoff; j < dst_xoff + dst_xsize; j++, k++ ) {
-            if(SKIP_MASK(j,i)) {
-              continue;
-            }
-            if( MS_VALID_COLOR( layer->offsite )
-                && pabyRaw1[k] == layer->offsite.red
-                && pabyRaw2[k] == layer->offsite.green
-                && pabyRaw3[k] == layer->offsite.blue )
-              continue;
-
-            if( bHaveRGBNoData
-                && pabyRaw1[k] == nNoData1
-                && pabyRaw2[k] == nNoData2
-                && pabyRaw3[k] == nNoData3 )
-              continue;
-
-            if( pabyRawAlpha != NULL && pabyRawAlpha[k] == 0 )
-              continue;
-
-            rb->data.gd_img->pixels[i][j] = pabyDithered[k];
-          }
-        }
-
-        free( pabyDithered );
-      }
-
-      /* Color cubed 24bit to 8bit conversion. */
-      else if( rb->type == MS_BUFFER_GD ) {
-        k = 0;
-        for( i = dst_yoff; i < dst_yoff + dst_ysize; i++ ) {
-          for( j = dst_xoff; j < dst_xoff + dst_xsize; j++, k++ ) {
-            int cc_index;
-            if(SKIP_MASK(j,i)) {
-              continue;
-            }
-
-            if( MS_VALID_COLOR( layer->offsite )
-                && pabyRaw1[k] == layer->offsite.red
-                && pabyRaw2[k] == layer->offsite.green
-                && pabyRaw3[k] == layer->offsite.blue )
-              continue;
-
-            if( bHaveRGBNoData
-                && pabyRaw1[k] == nNoData1
-                && pabyRaw2[k] == nNoData2
-                && pabyRaw3[k] == nNoData3 )
-              continue;
-
-            if( pabyRawAlpha != NULL && pabyRawAlpha[k] == 0 )
-              continue;
-
-            cc_index= RGB_INDEX(pabyRaw1[k],pabyRaw2[k],pabyRaw3[k]);
-            rb->data.gd_img->pixels[i][j] = anColorCube[cc_index];
-          }
-        }
-      } else {
-        msSetError(MS_MISCERR,"Unsupported raster configuration","msDrawRasterLayerGDAL()");
-        return MS_FAILURE;
-      }
-    }
-#endif
 
   /*
   ** Cleanup
@@ -1120,8 +889,8 @@ static int ParseDefaultLUT( const char *lut_def, GByte *lut )
         lut_read++;
     }
 
-    this_in = MAX(0,MIN(255,this_in));
-    this_out = MAX(0,MIN(255,this_out));
+    this_in = MS_MAX(0,MS_MIN(255,this_in));
+    this_out = MS_MAX(0,MS_MIN(255,this_out));
 
     /* apply linear values from last in:out to this in:out */
     for( lut_i = last_in; lut_i <= this_in; lut_i++ ) {
@@ -1506,8 +1275,8 @@ LoadGDALImages( GDALDatasetH hDS, int band_numbers[4], int band_count,
           bMinMaxSet = TRUE;
         }
 
-        dfScaleMin = MIN(dfScaleMin,pafRawData[i]);
-        dfScaleMax = MAX(dfScaleMax,pafRawData[i]);
+        dfScaleMin = MS_MIN(dfScaleMin,pafRawData[i]);
+        dfScaleMax = MS_MAX(dfScaleMax,pafRawData[i]);
       }
 
       if( dfScaleMin == dfScaleMax )
@@ -1563,152 +1332,6 @@ LoadGDALImages( GDALDatasetH hDS, int band_numbers[4], int band_count,
   return result_code;
 }
 
-#ifdef USE_GD
-/************************************************************************/
-/*                           allocColorCube()                           */
-/*                                                                      */
-/*      Allocate color table entries as best possible for a color       */
-/*      cube.                                                           */
-/************************************************************************/
-
-static int allocColorCube(mapObj *map, gdImagePtr img, int *panColorCube)
-
-{
-  int  r, g, b;
-  int i = 0;
-  int nGreyColors = 32;
-  int nSpaceGreyColors = 8;
-  int iColors = 0;
-  int red, green, blue;
-
-  for( r = 0; r < RED_LEVELS; r++ ) {
-    for( g = 0; g < GREEN_LEVELS; g++ ) {
-      for( b = 0; b < BLUE_LEVELS; b++ ) {
-        red = MS_MIN(255,r * (255 / (RED_LEVELS-1)));
-        green = MS_MIN(255,g * (255 / (GREEN_LEVELS-1)));
-        blue = MS_MIN(255,b * (255 / (BLUE_LEVELS-1)));
-
-        panColorCube[RGB_LEVEL_INDEX(r,g,b)] =
-          msAddColorGD(map, img, 1, red, green, blue );
-        iColors++;
-      }
-    }
-  }
-  /* -------------------------------------------------------------------- */
-  /*      Adding 32 grey colors                                           */
-  /* -------------------------------------------------------------------- */
-  for (i=0; i<nGreyColors; i++) {
-    red = i*nSpaceGreyColors;
-    green = red;
-    blue = red;
-    if(iColors < 256) {
-      panColorCube[iColors] = msAddColorGD(map,img,1,red,green,blue);
-      iColors++;
-    }
-  }
-  return MS_SUCCESS;
-}
-
-/************************************************************************/
-/*                            Dither24to8()                             */
-/*                                                                      */
-/*      Wrapper for GDAL dithering algorithm.                           */
-/************************************************************************/
-
-static void Dither24to8( GByte *pabyRed, GByte *pabyGreen, GByte *pabyBlue,
-                         GByte *pabyDithered, int xsize, int ysize,
-                         int bTransparent, colorObj transparent,
-                         gdImagePtr gdImg )
-
-{
-  GDALDatasetH hDS;
-  GDALDriverH  hDriver;
-  char **papszBandOptions = NULL;
-  char szDataPointer[120];
-  GDALColorTableH hCT;
-  int c;
-
-  /* -------------------------------------------------------------------- */
-  /*      Create the "memory" GDAL dataset referencing our working        */
-  /*      arrays.                                                         */
-  /* -------------------------------------------------------------------- */
-  hDriver = GDALGetDriverByName( "MEM" );
-  if( hDriver == NULL )
-    return;
-
-  hDS = GDALCreate( hDriver, "TempDitherDS", xsize, ysize, 0, GDT_Byte,
-                    NULL );
-
-  /* Add Red Band */
-  memset( szDataPointer, 0, sizeof(szDataPointer) );
-  CPLPrintPointer( szDataPointer, pabyRed, sizeof(szDataPointer) );
-  papszBandOptions = CSLSetNameValue( papszBandOptions, "DATAPOINTER",
-                                      szDataPointer );
-  GDALAddBand( hDS, GDT_Byte, papszBandOptions );
-
-  /* Add Green Band */
-  memset( szDataPointer, 0, sizeof(szDataPointer) );
-  CPLPrintPointer( szDataPointer, pabyGreen, sizeof(szDataPointer) );
-  papszBandOptions = CSLSetNameValue( papszBandOptions, "DATAPOINTER",
-                                      szDataPointer );
-  GDALAddBand( hDS, GDT_Byte, papszBandOptions );
-
-  /* Add Blue Band */
-  memset( szDataPointer, 0, sizeof(szDataPointer) );
-  CPLPrintPointer( szDataPointer, pabyBlue, sizeof(szDataPointer) );
-  papszBandOptions = CSLSetNameValue( papszBandOptions, "DATAPOINTER",
-                                      szDataPointer );
-  GDALAddBand( hDS, GDT_Byte, papszBandOptions );
-
-  /* Add Dithered Band */
-  memset( szDataPointer, 0, sizeof(szDataPointer) );
-  CPLPrintPointer( szDataPointer, pabyDithered, sizeof(szDataPointer) );
-  papszBandOptions = CSLSetNameValue( papszBandOptions, "DATAPOINTER",
-                                      szDataPointer );
-  GDALAddBand( hDS, GDT_Byte, papszBandOptions );
-
-  CSLDestroy( papszBandOptions );
-
-  /* -------------------------------------------------------------------- */
-  /*      Create the color table.                                         */
-  /* -------------------------------------------------------------------- */
-  hCT = GDALCreateColorTable( GPI_RGB );
-
-  for (c = 0; c < gdImg->colorsTotal; c++) {
-    GDALColorEntry sEntry;
-
-    sEntry.c1 = gdImg->red[c];
-    sEntry.c2 = gdImg->green[c];
-    sEntry.c3 = gdImg->blue[c];
-
-    if( bTransparent
-        && transparent.red == sEntry.c1
-        && transparent.green == sEntry.c2
-        && transparent.blue == sEntry.c3 )
-      sEntry.c4 = 0;
-    else
-      sEntry.c4 = 255;
-
-    GDALSetColorEntry( hCT, c, &sEntry );
-  }
-
-  /* -------------------------------------------------------------------- */
-  /*      Perform dithering.                                              */
-  /* -------------------------------------------------------------------- */
-  GDALDitherRGB2PCT( GDALGetRasterBand( hDS, 1 ),
-                     GDALGetRasterBand( hDS, 2 ),
-                     GDALGetRasterBand( hDS, 3 ),
-                     GDALGetRasterBand( hDS, 4 ),
-                     hCT, NULL, NULL );
-
-  /* -------------------------------------------------------------------- */
-  /*      Cleanup.                                                        */
-  /* -------------------------------------------------------------------- */
-  GDALDestroyColorTable( hCT );
-  GDALClose( hDS );
-}
-#endif
-
 
 /************************************************************************/
 /*                       msGetGDALGeoTransform()                        */
@@ -2100,7 +1723,7 @@ msDrawRasterLayerGDAL_16BitClassification(
       return -1;
   }
 
-  assert( rb->type == MS_BUFFER_GD || rb->type == MS_BUFFER_BYTE_RGBA );
+  assert( rb->type == MS_BUFFER_BYTE_RGBA );
 
   /* ==================================================================== */
   /*      Read the requested data in one gulp into a floating point       */
@@ -2145,8 +1768,8 @@ msDrawRasterLayerGDAL_16BitClassification(
       fDataMin = fDataMax = pafRawData[i];
       bGotFirstValue = TRUE;
     } else {
-      fDataMin = MIN(fDataMin,pafRawData[i]);
-      fDataMax = MAX(fDataMax,pafRawData[i]);
+      fDataMin = MS_MIN(fDataMin,pafRawData[i]);
+      fDataMax = MS_MAX(fDataMax,pafRawData[i]);
     }
   }
 
@@ -2261,30 +1884,17 @@ msDrawRasterLayerGDAL_16BitClassification(
       for(s=0; s<layer->class[c]->numstyles; s++) {
         if( MS_VALID_COLOR(layer->class[c]->styles[s]->mincolor)
             && MS_VALID_COLOR(layer->class[c]->styles[s]->maxcolor) )
-          msValueToRange(layer->class[c]->styles[s],dfOriginalValue);
+          msValueToRange(layer->class[c]->styles[s],dfOriginalValue, MS_COLORSPACE_RGB);
+      }
+      if( MS_TRANSPARENT_COLOR(layer->class[c]->styles[0]->color) ) {
+        /* leave it transparent */
+      } else if( MS_VALID_COLOR(layer->class[c]->styles[0]->color)) {
+        /* use class color */
+        rb_cmap[0][i] = layer->class[c]->styles[0]->color.red;
+        rb_cmap[1][i] = layer->class[c]->styles[0]->color.green;
+        rb_cmap[2][i] = layer->class[c]->styles[0]->color.blue;
+        rb_cmap[3][i] = (255*layer->class[c]->styles[0]->opacity / 100);
       }
-#ifdef USE_GD
-      if(rb->type == MS_BUFFER_GD) {
-        RESOLVE_PEN_GD(rb->data.gd_img, layer->class[c]->styles[0]->color);
-        if( MS_TRANSPARENT_COLOR(layer->class[c]->styles[0]->color) )
-          cmap[i] = -1;
-        else if( MS_VALID_COLOR(layer->class[c]->styles[0]->color)) {
-          /* use class color */
-          cmap[i] = layer->class[c]->styles[0]->color.pen;
-        }
-      } else
-#endif
-        if( rb->type == MS_BUFFER_BYTE_RGBA ) {
-          if( MS_TRANSPARENT_COLOR(layer->class[c]->styles[0]->color) ) {
-            /* leave it transparent */
-          } else if( MS_VALID_COLOR(layer->class[c]->styles[0]->color)) {
-            /* use class color */
-            rb_cmap[0][i] = layer->class[c]->styles[0]->color.red;
-            rb_cmap[1][i] = layer->class[c]->styles[0]->color.green;
-            rb_cmap[2][i] = layer->class[c]->styles[0]->color.blue;
-            rb_cmap[3][i] = (255*layer->class[c]->styles[0]->opacity / 100);
-          }
-        }
     }
   }
 
@@ -2317,24 +1927,13 @@ msDrawRasterLayerGDAL_16BitClassification(
       if( iMapIndex >= nBucketCount || iMapIndex < 0 ) {
         continue;
       }
-#ifdef USE_GD
-      if( rb->type == MS_BUFFER_GD ) {
-        int result = cmap[iMapIndex];
-        if( result == -1 )
-          continue;
-
-        rb->data.gd_img->pixels[i][j] = result;
-      } else
-#endif
-        if( rb->type == MS_BUFFER_BYTE_RGBA ) {
-          /* currently we never have partial alpha so keep simple */
-          if( rb_cmap[3][iMapIndex] > 0 )
-            RB_SET_PIXEL( rb, j, i,
-                          rb_cmap[0][iMapIndex],
-                          rb_cmap[1][iMapIndex],
-                          rb_cmap[2][iMapIndex],
-                          rb_cmap[3][iMapIndex] );
-        }
+      /* currently we never have partial alpha so keep simple */
+      if( rb_cmap[3][iMapIndex] > 0 )
+        RB_SET_PIXEL( rb, j, i,
+                      rb_cmap[0][iMapIndex],
+                      rb_cmap[1][iMapIndex],
+                      rb_cmap[2][iMapIndex],
+                      rb_cmap[3][iMapIndex] );
     }
   }
 
@@ -2419,7 +2018,7 @@ int *msGetGDALBandList( layerObj *layer, void *hDS,
   /* -------------------------------------------------------------------- */
   if( CSLFetchNameValue( layer->processing, "BANDS" ) == NULL ) {
     if( max_bands > 0 )
-      *band_count = MIN(file_bands,max_bands);
+      *band_count = MS_MIN(file_bands,max_bands);
     else
       *band_count = file_bands;
 
