@@ -31,6 +31,7 @@
 #include "mapserver.h"
 #include "mapproject.h"
 #include "mapthread.h"
+#include "mapows.h"
 
 #if defined(USE_OGR)
 #  define __USE_LARGEFILE64 1
@@ -86,6 +87,21 @@ int msInitDefaultOGROutputFormat( outputFormatObj *format )
   return MS_SUCCESS;
 }
 
+/************************************************************************/
+/*                        msCSLConcatenate()                            */
+/************************************************************************/
+
+static char** msCSLConcatenate( char** papszResult, char** papszToBeAdded )
+{
+    char** papszIter = papszToBeAdded;
+    while( papszIter && *papszIter )
+    {
+        papszResult = CSLAddString(papszResult, *papszIter);
+        papszIter ++;
+    }
+    return papszResult;
+}
+
 /************************************************************************/
 /*                       msOGRRecursiveFileList()                       */
 /*                                                                      */
@@ -139,7 +155,7 @@ char **msOGRRecursiveFileList( const char *path )
     } else if( VSI_ISDIR( sStatBuf.st_mode ) ) {
       char **subfiles = msOGRRecursiveFileList( full_filename );
 
-      result_list = CSLMerge( result_list, subfiles );
+      result_list = msCSLConcatenate( result_list, subfiles );
 
       CSLDestroy( subfiles );
     }
@@ -159,8 +175,15 @@ static void msOGRCleanupDS( const char *datasource_name )
   char **file_list;
   char path[MS_MAXPATHLEN];
   int i;
+  VSIStatBufL sStatBuf;
+  
+  if( VSIStatL( datasource_name, &sStatBuf ) != 0 )
+    return;
+  if( VSI_ISDIR( sStatBuf.st_mode ) )
+    strlcpy( path, datasource_name, sizeof(path) );
+  else
+    strlcpy( path, CPLGetPath( datasource_name ), sizeof(path) );
 
-  strlcpy( path, CPLGetPath( datasource_name ), sizeof(path) );
   file_list = CPLReadDir( path );
 
   for( i = 0; file_list != NULL && file_list[i] != NULL; i++ ) {
@@ -179,11 +202,7 @@ static void msOGRCleanupDS( const char *datasource_name )
     if( VSI_ISREG( sStatBuf.st_mode ) ) {
       VSIUnlink( full_filename );
     } else if( VSI_ISDIR( sStatBuf.st_mode ) ) {
-      char fake_ds_name[MS_MAXPATHLEN];
-      strlcpy( fake_ds_name,
-               CPLFormFilename( full_filename, "abc.dat", NULL ),
-               sizeof(fake_ds_name) );
-      msOGRCleanupDS( fake_ds_name );
+      msOGRCleanupDS( full_filename );
     }
   }
 
@@ -227,7 +246,8 @@ static void msOGRSetPoints( OGRGeometryH hGeom, lineObj *line, int bWant2DOutput
 /************************************************************************/
 
 static int msOGRWriteShape( layerObj *map_layer, OGRLayerH hOGRLayer,
-                            shapeObj *shape, gmlItemListObj *item_list )
+                            shapeObj *shape, gmlItemListObj *item_list,
+                            int nFirstOGRFieldIndex )
 
 {
   OGRGeometryH hGeom = NULL;
@@ -256,41 +276,71 @@ static int msOGRWriteShape( layerObj *map_layer, OGRLayerH hOGRLayer,
                  "msOGRWriteShape()");
       return MS_FAILURE;
     }
-
-    if( shape->numlines > 1 )
-      hMP = OGR_G_CreateGeometry( wkbMultiPoint );
-
-    for( j = 0; j < shape->numlines; j++ ) {
-      if( shape->line[j].numpoints != 1 ) {
-        msSetError(MS_MISCERR,
-                   "Failed on odd point geometry.",
-                   "msOGRWriteShape()");
-        return MS_FAILURE;
-      }
-
-      hGeom = OGR_G_CreateGeometry( wkbPoint );
-      if( bWant2DOutput ) {
-        OGR_G_SetPoint_2D( hGeom, 0,
-                           shape->line[j].point[0].x,
-                           shape->line[j].point[0].y );
-      }
-      else {
-        OGR_G_SetPoint( hGeom, 0,
-                        shape->line[j].point[0].x,
-                        shape->line[j].point[0].y,
+    
+    if( shape->numlines == 1 && shape->line[0].numpoints > 1 )
+    {
+      hGeom = OGR_G_CreateGeometry( wkbMultiPoint );
+      for( j = 0; j < shape->line[0].numpoints; j++ ) {
+        OGRGeometryH hPoint = OGR_G_CreateGeometry( wkbPoint );
+        if( bWant2DOutput ) {
+            OGR_G_SetPoint_2D( hPoint, 0,
+                            shape->line[0].point[j].x,
+                            shape->line[0].point[j].y );
+        }
+        else {
+            OGR_G_SetPoint( hPoint, 0,
+                            shape->line[0].point[j].x,
+                            shape->line[0].point[j].y,
 #ifdef USE_POINT_Z_M
-                        shape->line[j].point[0].z
+                            shape->line[0].point[j].z
 #else
-                        0.0
+                            0.0
 #endif
-                        );
-      }
+                            );
+        }
 
-      if( hMP != NULL ) {
-        OGR_G_AddGeometryDirectly( hMP, hGeom );
-        hGeom = hMP;
+        OGR_G_AddGeometryDirectly( hGeom, hPoint );
       }
     }
+    else
+    {
+      if( shape->numlines > 1 )
+        hMP = OGR_G_CreateGeometry( wkbMultiPoint );
+
+      for( j = 0; j < shape->numlines; j++ ) {
+        if( shape->line[j].numpoints != 1 ) {
+          msSetError(MS_MISCERR,
+                    "Failed on odd point geometry.",
+                    "msOGRWriteShape()");
+          return MS_FAILURE;
+        }
+
+        hGeom = OGR_G_CreateGeometry( wkbPoint );
+        if( bWant2DOutput ) {
+            OGR_G_SetPoint_2D( hGeom, 0,
+                            shape->line[j].point[0].x,
+                            shape->line[j].point[0].y );
+        }
+        else {
+            OGR_G_SetPoint( hGeom, 0,
+                            shape->line[j].point[0].x,
+                            shape->line[j].point[0].y,
+#ifdef USE_POINT_Z_M
+                            shape->line[j].point[0].z
+#else
+                            0.0
+#endif
+                            );
+        }
+
+        if( hMP != NULL ) {
+            OGR_G_AddGeometryDirectly( hMP, hGeom );
+        }
+      }
+      
+      if( hMP != NULL )
+        hGeom = hMP;
+    }
   }
 
   /* -------------------------------------------------------------------- */
@@ -435,7 +485,7 @@ static int msOGRWriteShape( layerObj *map_layer, OGRLayerH hOGRLayer,
   /* -------------------------------------------------------------------- */
   /*      Set attributes.                                                 */
   /* -------------------------------------------------------------------- */
-  out_field = 0;
+  out_field = nFirstOGRFieldIndex;
   for( i = 0; i < item_list->numitems; i++ ) {
     gmlItemObj *item = item_list->items + i;
 
@@ -447,7 +497,11 @@ static int msOGRWriteShape( layerObj *map_layer, OGRLayerH hOGRLayer,
     if( shape->values[i][0] == '\0' ) {
       OGRFieldDefnH hFieldDefn = OGR_FD_GetFieldDefn(hLayerDefn, out_field);
       OGRFieldType eFieldType = OGR_Fld_GetType(hFieldDefn);
-      if( eFieldType == OFTInteger || eFieldType == OFTReal )
+      if( eFieldType == OFTInteger || eFieldType == OFTReal
+#if GDAL_VERSION_MAJOR >= 2
+          || eFieldType == OFTInteger64
+#endif
+          )
       {
         out_field++;
         continue;
@@ -478,6 +532,133 @@ static int msOGRWriteShape( layerObj *map_layer, OGRLayerH hOGRLayer,
     return MS_FAILURE;
 }
 
+#if defined(GDAL_COMPUTE_VERSION)
+#if GDAL_VERSION_NUM >= GDAL_COMPUTE_VERSION(2,0,0)
+/************************************************************************/
+/*                        msOGRStdoutWriteFunction()                    */
+/************************************************************************/
+
+/* Used by /vsistdout/ */
+static size_t msOGRStdoutWriteFunction(const void* ptr, size_t size, size_t nmemb, FILE* stream)
+{
+    msIOContext *ioctx = (msIOContext*) stream;
+    return msIO_contextWrite(ioctx, ptr, size * nmemb ) / size;
+}
+#endif
+#endif
+
+/************************************************************************/
+/*                      msOGROutputGetAdditonalFiles()                  */
+/*                                                                      */
+/*  Collect additional files specified in                               */
+/*  wfs/ows_additional_files_in_output of WEB.METADATA and LAYER.METADATA */
+/************************************************************************/
+
+/* Result to be freed with CSLDestroy() */
+static char** msOGROutputGetAdditonalFiles( mapObj *map )
+{
+    int i;
+    hashTableObj* hSetAdditionalFiles;
+    char** papszFiles = NULL;
+
+    hSetAdditionalFiles = msCreateHashTable();
+
+    for( i = -1; i < map->numlayers; i++ )
+    {
+        const char* value;
+        if( i < 0 )
+        {
+            value = msOWSLookupMetadata(&(map->web.metadata), "FO", "additional_files_in_output");
+        }
+        else
+        {
+            layerObj *layer = GET_LAYER(map, i);
+            if( !layer->resultcache || layer->resultcache->numresults == 0 )
+                continue;
+            value = msOWSLookupMetadata(&(layer->metadata), "FO", "additional_files_in_output");
+        }
+
+        if( value != NULL )
+        {
+            char** papszList = CSLTokenizeString2( value, ",", CSLT_HONOURSTRINGS );
+            char** papszListIter = papszList;
+            while( papszListIter && *papszListIter )
+            {
+                const char* file = *papszListIter;
+                VSIStatBufL sStat;
+
+                if( strncmp(file, "http://", strlen("http://")) == 0 ||
+                    strncmp(file, "https://", strlen("https://")) == 0 )
+                {
+                    /* Remote file ? We will use /vsicurl_streaming/ to read it */
+                    if( msLookupHashTable(hSetAdditionalFiles, file) == NULL )
+                    {
+                        msInsertHashTable(hSetAdditionalFiles, file, "YES");
+                        papszFiles = CSLAddString(papszFiles, CPLSPrintf("/vsicurl_streaming/%s", file));
+                    }
+                }
+                else
+                {
+                    int nLen = (int)strlen(file);
+                    char filename[MS_MAXPATHLEN];
+
+                    if( CPLIsFilenameRelative(file) )
+                    {
+                        if( !map->shapepath )
+                            msTryBuildPath(filename, map->mappath, file);
+                        else
+                            msTryBuildPath3(filename, map->mappath, map->shapepath, file);
+                    }
+                    else
+                        strlcpy(filename, file, MS_MAXPATHLEN);
+
+                    if( nLen > 2 && (
+                            strcmp(file + nLen - 1, "/") == 0 ||
+                            strcmp(file + nLen - 2, "/*") == 0 ) )
+                    {
+                        *strrchr(filename, '/') = '\0';
+                    }
+                    else if( nLen > 2 && (
+                            strcmp(file + nLen - 1, "\\") == 0 ||
+                            strcmp(file + nLen - 2, "\\*") == 0 ) )
+                    {
+                        *strrchr(filename, '\\') = '\0';
+                    }
+
+                    if( msLookupHashTable(hSetAdditionalFiles, filename) == NULL )
+                    {
+                        msInsertHashTable(hSetAdditionalFiles, filename, "YES");
+                        if( VSIStatL( filename, &sStat ) == 0 )
+                        {
+                            if( VSI_ISDIR( sStat.st_mode ) )
+                            {
+                                char** papszDirContent = msOGRRecursiveFileList(filename);
+                                papszFiles = msCSLConcatenate(papszFiles, papszDirContent);
+                                CSLDestroy(papszDirContent);
+                            }
+                            else
+                            {
+                                papszFiles = CSLAddString(papszFiles, filename);
+                            }
+                        }
+                        else
+                        {
+                            msDebug("File %s does not exist.\n", filename);
+                        }
+                    }
+                }
+
+                papszListIter ++;
+            }
+            CSLDestroy(papszList);
+        }
+    }
+    
+    msFreeHashTable(hSetAdditionalFiles);
+    
+    return papszFiles;
+}
+
 #endif /* def USE_OGR */
 
 /************************************************************************/
@@ -507,6 +688,7 @@ int msOGRWriteFromQuery( mapObj *map, outputFormatObj *format, int sendheaders )
   char **layer_options = NULL;
   char **file_list = NULL;
   int iLayer, i;
+  int bDataSourceNameIsRequestDir = FALSE;
 
   /* -------------------------------------------------------------------- */
   /*      Fetch the output format driver.                                 */
@@ -536,6 +718,20 @@ int msOGRWriteFromQuery( mapObj *map, outputFormatObj *format, int sendheaders )
   /*      Determine the output datasource name to use.                    */
   /* ==================================================================== */
   storage = msGetOutputFormatOption( format, "STORAGE", "filesystem" );
+  if( EQUAL(storage,"stream") && !msIO_isStdContext() ) {
+#if defined(GDAL_COMPUTE_VERSION)
+#if GDAL_VERSION_NUM >= GDAL_COMPUTE_VERSION(2,0,0)
+    msIOContext *ioctx = msIO_getHandler (stdout);
+    if( ioctx != NULL )
+        VSIStdoutSetRedirection( msOGRStdoutWriteFunction, (FILE*)ioctx );
+    else
+#endif
+#endif
+    /* bug #4858, streaming output won't work if standard output has been
+     * redirected, we switch to memory output in this case
+     */
+    storage = "memory";
+  }
 
   /* -------------------------------------------------------------------- */
   /*      Where are we putting stuff?                                     */
@@ -579,7 +775,16 @@ int msOGRWriteFromQuery( mapObj *map, outputFormatObj *format, int sendheaders )
   /* -------------------------------------------------------------------- */
   /*      Setup the full datasource name.                                 */
   /* -------------------------------------------------------------------- */
-  fo_filename = msGetOutputFormatOption( format, "FILENAME", "result.dat" );
+#if !defined(CPL_ZIP_API_OFFERED)
+  form = msGetOutputFormatOption( format, "FORM", "multipart" );
+#else
+  form = msGetOutputFormatOption( format, "FORM", "zip" );
+#endif
+
+  if( EQUAL(form,"zip") )
+    fo_filename = msGetOutputFormatOption( format, "FILENAME", "result.zip" );
+  else
+    fo_filename = msGetOutputFormatOption( format, "FILENAME", "result.dat" );
 
   /* Validate that the filename does not contain any directory */
   /* information, which might lead to removal of unwanted files. (#4086) */
@@ -593,7 +798,32 @@ int msOGRWriteFromQuery( mapObj *map, outputFormatObj *format, int sendheaders )
   }
 
   if( !EQUAL(storage,"stream") )
+  {
     msBuildPath( datasource_name, request_dir, fo_filename );
+
+    if( EQUAL(form,"zip") )
+    {
+      /* if generating a zip file, remove the zip extension for the internal */
+      /* filename */
+      if( EQUAL(CPLGetExtension(datasource_name), "zip") ) {
+        *strrchr(datasource_name, '.') = '\0';
+      }
+
+      /* and add .dat extension if user didn't provide another extension */
+      if( EQUAL(CPLGetExtension(datasource_name), "") ) {
+        strcat(datasource_name, ".dat");
+      }
+    }
+
+    /* Shapefile and MapInfo driver only properly work with multiple layers */
+    /* if the output dataset name is a directory */
+    if( EQUAL(format->driver+4, "ESRI Shapefile") ||
+        EQUAL(format->driver+4, "MapInfo File") )
+    {
+        bDataSourceNameIsRequestDir = TRUE;
+        strcpy(datasource_name, request_dir);
+    }
+  }
   else
     strcpy( datasource_name, "/vsistdout/" );
 
@@ -641,8 +871,9 @@ int msOGRWriteFromQuery( mapObj *map, outputFormatObj *format, int sendheaders )
     const char *value;
     char *pszWKT;
     int  reproject = MS_FALSE;
+    int  nFirstOGRFieldIndex = -1;
 
-    if( !layer->resultcache || layer->resultcache->numresults == 0 )
+    if( !layer->resultcache )
       continue;
 
     /* -------------------------------------------------------------------- */
@@ -725,9 +956,9 @@ int msOGRWriteFromQuery( mapObj *map, outputFormatObj *format, int sendheaders )
       OGR_DS_Destroy( hDS );
       msOGRCleanupDS( datasource_name );
       msSetError( MS_MISCERR,
-                  "OGR CreateDataSource failed for '%s' with driver '%s'.",
+                  "OGR OGR_DS_CreateLayer failed for layer '%s' with driver '%s'.",
                   "msOGRWriteFromQuery()",
-                  datasource_name,
+                  layer->name,
                   format->driver+4 );
       return MS_FAILURE;
     }
@@ -760,6 +991,12 @@ int msOGRWriteFromQuery( mapObj *map, outputFormatObj *format, int sendheaders )
         eType = OFTString;
       else if( EQUAL(item->type,"Integer") )
         eType = OFTInteger;
+      else if( EQUAL(item->type,"Long") )
+#if GDAL_VERSION_MAJOR >= 2
+        eType = OFTInteger64;
+#else
+        eType = OFTReal;
+#endif
       else if( EQUAL(item->type,"Real") )
         eType = OFTReal;
       else if( EQUAL(item->type,"Character") )
@@ -792,6 +1029,11 @@ int msOGRWriteFromQuery( mapObj *map, outputFormatObj *format, int sendheaders )
         msOGRCleanupDS( datasource_name );
         return MS_FAILURE;
       }
+
+      /* The index of the first field we create is not necessarily 0 */
+      if( nFirstOGRFieldIndex < 0 )
+          nFirstOGRFieldIndex = OGR_FD_GetFieldCount(
+                                        OGR_L_GetLayerDefn( hOGRLayer ) ) - 1;
     }
 
     /* -------------------------------------------------------------------- */
@@ -804,6 +1046,7 @@ int msOGRWriteFromQuery( mapObj *map, outputFormatObj *format, int sendheaders )
         if(status != MS_SUCCESS) {
           OGR_DS_Destroy( hDS );
           msOGRCleanupDS( datasource_name );
+          msGMLFreeItems(item_list);
           return status;
         }
       }
@@ -825,6 +1068,8 @@ int msOGRWriteFromQuery( mapObj *map, outputFormatObj *format, int sendheaders )
       if(status != MS_SUCCESS) {
         OGR_DS_Destroy( hDS );
         msOGRCleanupDS( datasource_name );
+        msGMLFreeItems(item_list);
+        msFreeShape(&resultshape);
         return status;
       }
 
@@ -839,8 +1084,7 @@ int msOGRWriteFromQuery( mapObj *map, outputFormatObj *format, int sendheaders )
               || layer->labelitem)
           && layer->class[resultshape.classindex]->numlabels > 0
           && layer->class[resultshape.classindex]->labels[0]->size != -1 ) {
-        msShapeGetAnnotation(layer, &resultshape); /* TODO RFC77: check return value */
-        resultshape.text = msStrdup(layer->class[resultshape.classindex]->labels[0]->annotext);
+        resultshape.text = msShapeGetLabelAnnotation(layer,&resultshape,layer->class[resultshape.classindex]->labels[0]);
       }
 
       /*
@@ -869,11 +1113,13 @@ int msOGRWriteFromQuery( mapObj *map, outputFormatObj *format, int sendheaders )
 
       if( status == MS_SUCCESS )
         status = msOGRWriteShape( layer, hOGRLayer, &resultshape,
-                                  item_list );
+                                  item_list, nFirstOGRFieldIndex );
 
       if(status != MS_SUCCESS) {
         OGR_DS_Destroy( hDS );
         msOGRCleanupDS( datasource_name );
+        msGMLFreeItems(item_list);
+        msFreeShape(&resultshape);
         return status;
       }
     }
@@ -890,19 +1136,19 @@ int msOGRWriteFromQuery( mapObj *map, outputFormatObj *format, int sendheaders )
   /* -------------------------------------------------------------------- */
   /*      Get list of resulting files.                                    */
   /* -------------------------------------------------------------------- */
-#if !defined(CPL_ZIP_API_OFFERED)
-  form = msGetOutputFormatOption( format, "FORM", "multipart" );
-#else
-  form = msGetOutputFormatOption( format, "FORM", "zip" );
-#endif
 
   if( EQUAL(form,"simple") ) {
     file_list = CSLAddString( NULL, datasource_name );
   } else {
     char datasource_path[MS_MAXPATHLEN];
 
-    strcpy( datasource_path, CPLGetPath( datasource_name ) );
-    file_list = msOGRRecursiveFileList( datasource_path );
+    if( bDataSourceNameIsRequestDir )
+        file_list = msOGRRecursiveFileList( datasource_name );
+    else
+    {
+        strncpy( datasource_path, CPLGetPath( datasource_name ), MS_MAXPATHLEN-1 );
+        file_list = msOGRRecursiveFileList( datasource_path );
+    }
   }
 
   /* -------------------------------------------------------------------- */
@@ -920,8 +1166,11 @@ int msOGRWriteFromQuery( mapObj *map, outputFormatObj *format, int sendheaders )
     char buffer[1024];
     int  bytes_read;
     FILE *fp;
+    const char *jsonp;
 
+    jsonp = msGetOutputFormatOption( format, "JSONP", NULL );
     if( sendheaders ) {
+      if( !jsonp )
       msIO_setHeader("Content-Disposition","attachment; filename=%s",
                      CPLGetFilename( file_list[0] ) );
       if( format->mimetype )
@@ -940,20 +1189,29 @@ int msOGRWriteFromQuery( mapObj *map, outputFormatObj *format, int sendheaders )
       return MS_FAILURE;
     }
 
+    if( jsonp != NULL ) msIO_fprintf( stdout, "%s(", jsonp );
+
     while( (bytes_read = VSIFReadL( buffer, 1, sizeof(buffer), fp )) > 0 )
       msIO_fwrite( buffer, 1, bytes_read, stdout );
     VSIFCloseL( fp );
+
+    if (jsonp != NULL) msIO_fprintf( stdout, ");\n" );
   }
 
   /* -------------------------------------------------------------------- */
   /*      Handle the case of a multi-part result.                         */
   /* -------------------------------------------------------------------- */
   else if( EQUAL(form,"multipart") ) {
+    char **papszAdditionalFiles;
     static const char *boundary = "xxOGRBoundaryxx";
     msIO_setHeader("Content-Type","multipart/mixed; boundary=%s",boundary);
     msIO_sendHeaders();
     msIO_fprintf(stdout,"--%s\r\n",boundary );
 
+    papszAdditionalFiles = msOGROutputGetAdditonalFiles(map);
+    file_list = msCSLConcatenate(file_list, papszAdditionalFiles);
+    CSLDestroy(papszAdditionalFiles);
+
     for( i = 0; file_list != NULL && file_list[i] != NULL; i++ ) {
       FILE *fp;
       int bytes_read;
@@ -1003,9 +1261,14 @@ int msOGRWriteFromQuery( mapObj *map, outputFormatObj *format, int sendheaders )
     void *hZip;
     int bytes_read;
     char buffer[1024];
+    char **papszAdditionalFiles;
 
     hZip = CPLCreateZip( zip_filename, NULL );
 
+    papszAdditionalFiles = msOGROutputGetAdditonalFiles(map);
+    file_list = msCSLConcatenate(file_list, papszAdditionalFiles);
+    CSLDestroy(papszAdditionalFiles);
+
     for( i = 0; file_list != NULL && file_list[i] != NULL; i++ ) {
 
       CPLCreateFileInZip( hZip, CPLGetFilename(file_list[i]), NULL );
@@ -1031,7 +1294,12 @@ int msOGRWriteFromQuery( mapObj *map, outputFormatObj *format, int sendheaders )
     CPLCloseZip( hZip );
 
     if( sendheaders ) {
-      msIO_setHeader("Content-Disposition","attachment; filename=%s",fo_filename);
+      const char* zip_filename = fo_filename;
+      /* Make sure the filename is ended by .zip */
+      if( !EQUAL(CPLGetExtension(zip_filename), "zip") &&
+          !EQUAL(CPLGetExtension(zip_filename), "kmz") )
+          zip_filename = CPLFormFilename(NULL, fo_filename, "zip");
+      msIO_setHeader("Content-Disposition","attachment; filename=%s",zip_filename);
       msIO_setHeader("Content-Type","application/zip");
       msIO_sendHeaders();
     }
