@@ -135,7 +135,7 @@ char *strrstr(char *string, char *find)
  * Appends src to string dst of size siz (unlike strncat, siz is the
  * full size of dst, not space left).  At most siz-1 characters
  * will be copied.  Always NUL terminates (unless siz <= strlen(dst)).
- * Returns strlen(src) + MIN(siz, strlen(initial dst)).
+ * Returns strlen(src) + MS_MIN(siz, strlen(initial dst)).
  * If retval >= siz, truncation occurred.
  */
 size_t strlcat(char *dst, const char *src, size_t siz)
@@ -414,6 +414,51 @@ void msStringToLower(char *string)
   }
 }
 
+/**
+ * Force the first character to uppercase and the rest of the characters to
+ * lower case for EACH word in the string.
+ */
+void msStringInitCap(char *string)
+{
+  int i;
+  int start = 1; 
+  if (string != NULL) {
+    for (i = 0; i < (int)strlen(string); i++) {
+      if (string[i] == ' ')
+        start = 1;
+      else if (start) {
+        string[i] = toupper(string[i]);
+        start = 0;
+      }
+      else {
+        string[i] = tolower(string[i]);
+      }
+    }
+  }
+}
+
+/**
+ * Force the first character to uppercase for the FIRST word in the string
+ * and the rest of the characters to lower case.
+ */
+void msStringFirstCap(char *string)
+{
+  int i;
+  int start = 1; 
+  if (string != NULL) {
+    for (i = 0; i < (int)strlen(string); i++) {
+      if (string[i] != ' ') {
+        if (start) {
+          string[i] = toupper(string[i]);
+          start = 0;
+        }
+        else
+          string[i] = tolower(string[i]);
+      }
+    }
+  }
+}
+
 char *msStringChop(char *string)
 {
   int n;
@@ -789,11 +834,9 @@ char **msStringSplit(const char *string, char ch, int *num_tokens)
   }
 
   token = (char **) msSmallMalloc(sizeof(char *)*n);
-  if(!token) return(NULL);
 
   k = 0;
   token[k] = (char *)msSmallMalloc(sizeof(char)*(length+1));
-  if(!token[k]) return(NULL);
 
   j = 0;
   last_ch='\0';
@@ -807,7 +850,6 @@ char **msStringSplit(const char *string, char ch, int *num_tokens)
 
       k++;
       token[k] = (char *)msSmallMalloc(sizeof(char)*(length+1));
-      if(!token[k]) return(NULL);
 
       j = 0;
     } else {
@@ -859,7 +901,7 @@ char ** msStringSplitComplex( const char * pszString,
   int         bStripLeadSpaces = (nFlags & MS_STRIPLEADSPACES);
   int         bStripEndSpaces = (nFlags & MS_STRIPENDSPACES);
 
-  pszToken = (char *) msSmallMalloc(sizeof(char*)*10);;
+  pszToken = (char *) msSmallMalloc(sizeof(char)*10);;
   nTokenMax = 10;
 
   while( pszString != NULL && *pszString != '\0' ) {
@@ -925,7 +967,7 @@ char ** msStringSplitComplex( const char * pszString,
        */
       if( nTokenLen >= nTokenMax-3 ) {
         nTokenMax = nTokenMax * 2 + 10;
-        pszToken = (char *) msSmallRealloc(pszToken, sizeof(char*)*nTokenMax);
+        pszToken = (char *) msSmallRealloc(pszToken, sizeof(char)*nTokenMax);
       }
 
       pszToken[nTokenLen] = *pszString;
@@ -987,7 +1029,7 @@ char **msStringTokenize( const char *pszLine, const char *pszDelim,
 {
   char **papszResult = NULL;
   int n = 1, iChar, nLength = strlen(pszLine), iTokenChar = 0, bInQuotes = MS_FALSE;
-  char *pszToken = (char *) msSmallMalloc(sizeof(char*)*(nLength+1));
+  char *pszToken = (char *) msSmallMalloc(sizeof(char)*(nLength+1));
   int nDelimLen = strlen(pszDelim);
 
   /* Compute the number of tokens */
@@ -1017,7 +1059,7 @@ char **msStringTokenize( const char *pszLine, const char *pszDelim,
     } else if( !bInQuotes && strncmp(pszLine+iChar,pszDelim,nDelimLen) == 0 ) {
       pszToken[iTokenChar++] = '\0';
       papszResult[n] = pszToken;
-      pszToken = (char *) msSmallMalloc(sizeof(char*)*(nLength+1));
+      pszToken = (char *) msSmallMalloc(sizeof(char)*(nLength+1));
       iChar += nDelimLen - 1;
       iTokenChar = 0;
       n++;
@@ -1080,7 +1122,7 @@ char *msEncodeUrl(const char *data)
 
 char *msEncodeUrlExcept(const char *data, const char except)
 {
-  char *hex = "0123456789ABCDEF";
+  static const char *hex = "0123456789ABCDEF";
   const char *i;
   char  *j, *code;
   int   inc;
@@ -1093,9 +1135,7 @@ char *msEncodeUrlExcept(const char *data, const char except)
   code = (char*)msSmallMalloc(strlen(data)+inc+1);
 
   for (j=code, i=data; *i!='\0'; i++, j++) {
-    if (*i == ' ')
-      *j = '+';
-    else if ( except != '\0' && *i == except ) {
+    if ( except != '\0' && *i == except ) {
       *j = except;
     } else if (msEncodeChar(*i)) {
       ch = *i;
@@ -1110,6 +1150,77 @@ char *msEncodeUrlExcept(const char *data, const char except)
   return code;
 }
 
+/************************************************************************/
+/*                            msEscapeJSonString()                      */
+/************************************************************************/
+
+/* The input (and output) string are not supposed to start/end with double */
+/* quote characters. It is the responsibility of the caller to do that. */
+char* msEscapeJSonString(const char* pszJSonString)
+{
+    /* Worst case is one character to become \uABCD so 6 characters */
+    char* pszRet;
+    int i = 0, j = 0;
+    static const char* pszHex = "0123456789ABCDEF";
+    
+    pszRet = (char*) msSmallMalloc(strlen(pszJSonString) * 6 + 1);
+    /* From http://www.json.org/ */
+    for(i = 0; pszJSonString[i] != '\0'; i++)
+    {
+        unsigned char ch = pszJSonString[i];
+        if( ch == '\b' )
+        {
+            pszRet[j++] = '\\';
+            pszRet[j++] = 'b';
+        }
+        else if( ch == '\f' )
+        {
+            pszRet[j++] = '\\';
+            pszRet[j++] = 'f';
+        }
+        else if( ch == '\n' )
+        {
+            pszRet[j++] = '\\';
+            pszRet[j++] = 'n';
+        }
+        else if( ch == '\r' )
+        {
+            pszRet[j++] = '\\';
+            pszRet[j++] = 'r';
+        }
+        else if( ch == '\t' )
+        {
+            pszRet[j++] = '\\';
+            pszRet[j++] = 't';
+        }
+        else if( ch < 32 )
+        {
+            pszRet[j++] = '\\';
+            pszRet[j++] = 'u';
+            pszRet[j++] = '0';
+            pszRet[j++] = '0';
+            pszRet[j++] = pszHex[ch / 16];
+            pszRet[j++] = pszHex[ch % 16];
+        }
+        else if( ch == '"' )
+        {
+            pszRet[j++] = '\\';
+            pszRet[j++] = '"';
+        }
+        else if( ch == '\\' )
+        {
+            pszRet[j++] = '\\';
+            pszRet[j++] = '\\';
+        }
+        else
+        {
+            pszRet[j++] = ch;
+        }
+    }
+    pszRet[j] = '\0';
+    return pszRet;
+}
+
 /* msEncodeHTMLEntities()
 **
 ** Return a copy of string after replacing some problematic chars with their
@@ -1194,8 +1305,8 @@ void msDecodeHTMLEntities(const char *string)
     pszBuffer = (char*)string;
 
   bufferSize = strlen(pszBuffer);
-  pszReplace = (char*) msSmallMalloc(bufferSize);
-  pszEnd = (char*) msSmallMalloc(bufferSize);
+  pszReplace = (char*) msSmallMalloc(bufferSize+1);
+  pszEnd = (char*) msSmallMalloc(bufferSize+1);
 
   while((pszAmp = strchr(pszBuffer, '&')) != NULL) {
     /* Get the &...; */
@@ -1414,14 +1525,15 @@ char *msCaseReplaceSubstring(char *str, const char *old, const char *new)
   size_t str_len, old_len, new_len, tmp_offset;
   char *tmp_ptr;
 
-  if(new == NULL)
-    new = "";
-
   /*
   ** If old is not found then leave str alone
   */
   if( (tmp_ptr = (char *) strcasestr(str, old)) == NULL)
     return(str);
+  
+  if(new == NULL)
+    new = "";
+
 
   /*
   ** Grab some info about incoming strings
@@ -1606,6 +1718,7 @@ char *msGetEncodedString(const char *string, const char *encoding)
   const char *inp;
   char *outp, *out = NULL;
   size_t len, bufsize, bufleft, iconv_status;
+  assert(encoding);
 
 #ifdef USE_FRIBIDI
   msAcquireLock(TLOCK_FRIBIDI);
@@ -1618,7 +1731,7 @@ char *msGetEncodedString(const char *string, const char *encoding)
 #endif
   len = strlen(string);
 
-  if (len == 0 || (encoding && strcasecmp(encoding, "UTF-8")==0))
+  if (len == 0 || strcasecmp(encoding, "UTF-8")==0)
     return msStrdup(string);    /* Nothing to do: string already in UTF-8 */
 
   cd = iconv_open("UTF-8", encoding);
@@ -1778,7 +1891,8 @@ char* msConvertWideStringToUTF8 (const wchar_t* string, const char* encoding)
 int msGetNextGlyph(const char **in_ptr, char *out_string)
 {
   unsigned char in;
-  int numbytes=0,unicode;
+  int numbytes=0;
+  unsigned int unicode;
   int i;
 
   in = (unsigned char)**in_ptr;
@@ -1905,7 +2019,7 @@ static int cmp_entities(const void *e1, const void *e2)
  * - if the string does start with such entity,it returns the number of
  * bytes occupied by said entity, and stores the unicode value in *unicode
  */
-int msGetUnicodeEntity(const char *inptr, int *unicode)
+int msGetUnicodeEntity(const char *inptr, unsigned int *unicode)
 {
   unsigned char *in = (unsigned char*)inptr;
   int l,val=0;
@@ -2073,3 +2187,59 @@ int msStringInArray( const char * pszString, char **array, int numelements)
   }
   return MS_FALSE;
 }
+
+int msLayerEncodeShapeAttributes( layerObj *layer, shapeObj *shape) {
+
+#ifdef USE_ICONV
+  iconv_t cd = NULL;
+  const char *inp;
+  char *outp, *out = NULL;
+  size_t len, bufsize, bufleft, iconv_status;
+  int i;
+
+  if( !layer->encoding || !*layer->encoding || !strcasecmp(layer->encoding, "UTF-8"))
+    return MS_SUCCESS;
+
+  cd = iconv_open("UTF-8", layer->encoding);
+  if(cd == (iconv_t)-1) {
+    msSetError(MS_IDENTERR, "Encoding not supported by libiconv (%s).",
+               "msGetEncodedString()", layer->encoding);
+    return MS_FAILURE;
+  }
+
+  for(i=0;i <shape->numvalues; i++) {
+    if(!shape->values[i] || (len = strlen(shape->values[i]))==0) {
+      continue;    /* Nothing to do */
+    }
+
+    bufsize = len * 6 + 1; /* Each UTF-8 char can be up to 6 bytes */
+    inp = shape->values[i];
+    out = (char*) msSmallMalloc(bufsize);
+
+    strlcpy(out, shape->values[i], bufsize);
+    outp = out;
+
+    bufleft = bufsize;
+    iconv_status = -1;
+
+    while (len > 0) {
+      iconv_status = iconv(cd, (char**)&inp, &len, &outp, &bufleft);
+      if(iconv_status == -1) {
+        msFree(out);
+        continue; /* silently ignore failed conversions */
+      }
+    }
+    out[bufsize - bufleft] = '\0';
+    msFree(shape->values[i]);
+    shape->values[i] = out;
+  }
+  iconv_close(cd);
+
+  return MS_SUCCESS;
+#else
+  if( !layer->encoding || !*layer->encoding || !strcasecmp(layer->encoding, "UTF-8"))
+    return MS_SUCCESS;
+  msSetError(MS_MISCERR, "Not implemented since Iconv is not enabled.", "msGetEncodedString()");
+  return MS_FAILURE;
+#endif
+}
