@@ -37,6 +37,7 @@
 #include "maptime.h"
 #include "mapthread.h"
 #include "mapcopy.h"
+#include "mapows.h"
 
 #if defined(_WIN32) && !defined(__CYGWIN__)
 # include <windows.h>
@@ -216,7 +217,7 @@ static void bindLabel(layerObj *layer, shapeObj *shape, labelObj *label, int dra
 
     if(label->bindings[MS_LABEL_BINDING_SIZE].index != -1) {
       label->size = 1;
-      bindDoubleAttribute(&label->size, shape->values[label->bindings[MS_LABEL_BINDING_SIZE].index]);
+      bindIntegerAttribute(&label->size, shape->values[label->bindings[MS_LABEL_BINDING_SIZE].index]);
     }
 
     if(label->bindings[MS_LABEL_BINDING_COLOR].index != -1) {
@@ -403,7 +404,7 @@ int msEvalContext(mapObj *map, layerObj *layer, char *context)
   if(!context) return(MS_TRUE);
 
   /* initialize a temporary expression (e) */
-  initExpression(&e);
+  msInitExpression(&e);
 
   e.string = msStrdup(context);
   e.type = MS_EXPRESSION; /* todo */
@@ -434,7 +435,7 @@ int msEvalContext(mapObj *map, layerObj *layer, char *context)
 
   status = yyparse(&p);
 
-  freeExpression(&e);
+  msFreeExpression(&e);
 
   if (status != 0) {
     msSetError(MS_PARSEERR, "Failed to parse context", "msEvalContext");
@@ -454,7 +455,8 @@ int msEvalContext(mapObj *map, layerObj *layer, char *context)
  */
 int msEvalExpression(layerObj *layer, shapeObj *shape, expressionObj *expression, int itemindex)
 {
-  if(!expression->string) return MS_TRUE; /* empty expressions are ALWAYS true */
+  if(MS_STRING_IS_NULL_OR_EMPTY(expression->string)) return MS_TRUE; /* NULL or empty expressions are ALWAYS true */
+  if(expression->native_string != NULL) return MS_TRUE; /* expressions that are evaluated natively are ALWAYS true */
 
   switch(expression->type) {
     case(MS_STRING):
@@ -483,9 +485,10 @@ int msEvalExpression(layerObj *layer, shapeObj *shape, expressionObj *expression
       }
       {
         char *start,*end;
+        int value_len = strlen(shape->values[itemindex]);
         start = expression->string;
         while((end = strchr(start,',')) != NULL) {
-          if(!strncmp(start,shape->values[itemindex],end-start)) return MS_TRUE;
+          if(value_len == end-start && !strncmp(start,shape->values[itemindex],end-start)) return MS_TRUE;
           start = end+1;
         }
         if(!strcmp(start,shape->values[itemindex])) return MS_TRUE;
@@ -520,6 +523,8 @@ int msEvalExpression(layerObj *layer, shapeObj *shape, expressionObj *expression
         return MS_FALSE;
       }
 
+      if(MS_STRING_IS_NULL_OR_EMPTY(shape->values[itemindex]) == MS_TRUE) return MS_FALSE;
+
       if(!expression->compiled) {
         if(expression->flags & MS_EXP_INSENSITIVE) {
           if(ms_regcomp(&(expression->regex), expression->string, MS_REG_EXTENDED|MS_REG_NOSUB|MS_REG_ICASE) != 0) { /* compile the expression */
@@ -611,7 +616,8 @@ int msShapeGetClass(layerObj *layer, mapObj *map, shapeObj *shape, int *classgro
   return(-1); /* no match */
 }
 
-static char *evalTextExpression(expressionObj *expr, shapeObj *shape)
+static
+char *msEvalTextExpressionInternal(expressionObj *expr, shapeObj *shape, int bJSonEscape)
 {
   char *result=NULL;
 
@@ -620,6 +626,7 @@ static char *evalTextExpression(expressionObj *expr, shapeObj *shape)
   switch(expr->type) {
     case(MS_STRING): {
       char *target=NULL;
+      char *pszEscaped;
       tokenListNodeObjPtr node=NULL;
       tokenListNodeObjPtr nextNode=NULL;
 
@@ -632,7 +639,12 @@ static char *evalTextExpression(expressionObj *expr, shapeObj *shape)
           if(node->token == MS_TOKEN_BINDING_DOUBLE || node->token == MS_TOKEN_BINDING_INTEGER || node->token == MS_TOKEN_BINDING_STRING || node->token == MS_TOKEN_BINDING_TIME) {
             target = (char *) msSmallMalloc(strlen(node->tokenval.bindval.item) + 3);
             sprintf(target, "[%s]", node->tokenval.bindval.item);
-            result = msReplaceSubstring(result, target, shape->values[node->tokenval.bindval.index]);
+            if( bJSonEscape )
+                pszEscaped = msEscapeJSonString(shape->values[node->tokenval.bindval.index]);
+            else
+                pszEscaped = msStrdup(shape->values[node->tokenval.bindval.index]);
+            result = msReplaceSubstring(result, target, pszEscaped);
+            msFree(pszEscaped);
             msFree(target);
           }
           node = nextNode;
@@ -656,7 +668,7 @@ static char *evalTextExpression(expressionObj *expr, shapeObj *shape)
       status = yyparse(&p);
 
       if (status != 0) {
-        msSetError(MS_PARSEERR, "Failed to process text expression: %s", "evalTextExpression", expr->string);
+        msSetError(MS_PARSEERR, "Failed to process text expression: %s", "msEvalTextExpression", expr->string);
         return NULL;
       }
 
@@ -673,55 +685,39 @@ static char *evalTextExpression(expressionObj *expr, shapeObj *shape)
   return result;
 }
 
-int msShapeGetAnnotation(layerObj *layer, shapeObj *shape)
+char *msEvalTextExpressionJSonEscape(expressionObj *expr, shapeObj *shape)
 {
-  int i, j;
+    return msEvalTextExpressionInternal(expr, shape, MS_TRUE);
+}
 
-  /* RFC77 TODO: check and throw some errors here... */
-  if(!layer || !shape) return MS_FAILURE;
+char *msEvalTextExpression(expressionObj *expr, shapeObj *shape)
+{
+    return msEvalTextExpressionInternal(expr, shape, MS_FALSE);
+}
 
-  i = shape->classindex;
-  for(j=0; j<layer->class[i]->numlabels; j++) {
-    labelObj *lbl = layer->class[i]->labels[j]; /* shortcut */
-
-    lbl->status = MS_ON;
-    if(layer->map->scaledenom > 0) {
-      if((lbl->maxscaledenom != -1) && (layer->map->scaledenom >= lbl->maxscaledenom)) {
-        lbl->status = MS_OFF;
-        continue; /* next label */
-      }
-      if((lbl->minscaledenom != -1) && (layer->map->scaledenom < lbl->minscaledenom)) {
-        lbl->status = MS_OFF;
-        continue; /* next label */
-      }
-    }
-    if(msEvalExpression(layer, shape, &(lbl->expression), layer->labelitemindex) != MS_TRUE) {
-      lbl->status = MS_OFF;
-      continue; /* next label */
-    }
-
-    msFree(lbl->annotext);
-    lbl->annotext = NULL;
-
-    if(lbl->text.string) {
-      lbl->annotext = evalTextExpression(&(lbl->text), shape);
-    } else if(layer->class[i]->text.string) {
-      lbl->annotext = evalTextExpression(&(layer->class[i]->text), shape);
-    } else {
-      if (shape->values && layer->labelitemindex >= 0 && shape->values[layer->labelitemindex] && strlen(shape->values[layer->labelitemindex]) )
-        lbl->annotext = msStrdup(shape->values[layer->labelitemindex]);
-      else if(shape->text)
-        lbl->annotext = msStrdup(shape->text); /* last resort but common with iniline features */
-    }
-
-    if(lbl->annotext && (lbl->encoding || lbl->wrap || lbl->maxlength)) {
-      char *newtext = msTransformLabelText(layer->map , lbl, lbl->annotext);
-      free(lbl->annotext);
-      lbl->annotext = newtext;
-    }
+char* msShapeGetLabelAnnotation(layerObj *layer, shapeObj *shape, labelObj *lbl) {
+  assert(shape && lbl);
+  if(lbl->text.string) {
+    return msEvalTextExpression(&(lbl->text), shape);
+  } else if(layer->class[shape->classindex]->text.string) {
+    return msEvalTextExpression(&(layer->class[shape->classindex]->text), shape);
+  } else {
+    if (shape->values && layer->labelitemindex >= 0 && shape->values[layer->labelitemindex] && strlen(shape->values[layer->labelitemindex]) )
+      return msStrdup(shape->values[layer->labelitemindex]);
+    else if(shape->text)
+      return msStrdup(shape->text); /* last resort but common with iniline features */
   }
+  return NULL;
+}
 
-  return MS_SUCCESS;
+int msGetLabelStatus(mapObj *map, layerObj *layer, shapeObj *shape, labelObj *lbl) {
+  assert(layer && lbl);
+  if(!msScaleInBounds(map->scaledenom,lbl->minscaledenom,lbl->maxscaledenom))
+    return MS_OFF;
+  if(msEvalExpression(layer, shape, &(lbl->expression), layer->labelitemindex) != MS_TRUE)
+    return MS_OFF;
+  /* TODO: check for minfeaturesize here ? */
+  return MS_ON;
 }
 
 /* Check if the shape is enough big to be drawn with the
@@ -851,6 +847,7 @@ int msSaveImage(mapObj *map, imageObj *img, char *filename)
         nReturnVal = msSaveImageGDAL(map, img, filename);
     } else
 #endif
+
       if (MS_RENDERER_PLUGIN(img->format)) {
         rendererVTableObj *renderer = img->format->vtable;
         FILE *stream = NULL;
@@ -875,8 +872,11 @@ int msSaveImage(mapObj *map, imageObj *img, char *filename)
 
         if(renderer->supports_pixel_buffer) {
           rasterBufferObj data;
-          if(renderer->getRasterBufferHandle(img,&data) != MS_SUCCESS)
+          if(renderer->getRasterBufferHandle(img,&data) != MS_SUCCESS) {
+            if( stream != stdout )
+              fclose(stream);
             return MS_FAILURE;
+          }
 
           nReturnVal = msSaveRasterBuffer(map,&data,stream,img->format );
         } else {
@@ -915,6 +915,7 @@ int msSaveImage(mapObj *map, imageObj *img, char *filename)
 
 unsigned char *msSaveImageBuffer(imageObj* image, int *size_ptr, outputFormatObj *format)
 {
+  int status = MS_SUCCESS;
   *size_ptr = 0;
   if( MS_RENDERER_PLUGIN(image->format)) {
     rasterBufferObj data;
@@ -922,7 +923,10 @@ unsigned char *msSaveImageBuffer(imageObj* image, int *size_ptr, outputFormatObj
     if(renderer->supports_pixel_buffer) {
       bufferObj buffer;
       msBufferInit(&buffer);
-      renderer->getRasterBufferHandle(image,&data);
+      status = renderer->getRasterBufferHandle(image,&data);
+      if(UNLIKELY(status == MS_FAILURE)) {
+        return NULL;
+      }
       msSaveRasterBufferToBuffer(&data,&buffer,format);
       *size_ptr = buffer.size;
       return buffer.data;
@@ -1331,6 +1335,7 @@ char **msGetAllGroupNames(mapObj *map, int *numTok)
   int         nCount = 0;
   int         i = 0, j = 0;
 
+  assert(map);
   *numTok = 0;
 
   if (!map->layerorder) {
@@ -1343,7 +1348,7 @@ char **msGetAllGroupNames(mapObj *map, int *numTok)
       map->layerorder[i] = i;
   }
 
-  if (map != NULL && map->numlayers > 0) {
+  if (map->numlayers > 0) {
     nCount = map->numlayers;
     papszGroups = (char **)msSmallMalloc(sizeof(char *)*nCount);
 
@@ -1622,7 +1627,7 @@ imageObj *msImageCreate(int width, int height, outputFormatObj *format,
   }
 
   if(!image)
-    msSetError(MS_GDERR, "Unable to initialize image.", "msImageCreate()");
+    msSetError(MS_IMGERR, "Unable to initialize image.", "msImageCreate()");
   image->refpt.x = image->refpt.y = 0;
   return image;
 }
@@ -1640,10 +1645,6 @@ void  msTransformPoint(pointObj *point, rectObj *extent, double cellsize,
   if (image != NULL && MS_RENDERER_PLUGIN(image->format)) {
     if(image->format->renderer == MS_RENDER_WITH_KML) {
       return;
-    } else if(image->format->renderer == MS_RENDER_WITH_GD) {
-      point->x = MS_MAP2IMAGE_X(point->x, extent->minx, cellsize);
-      point->y = MS_MAP2IMAGE_Y(point->y, extent->maxy, cellsize);
-      return;
     }
   }
   invcellsize = 1.0/cellsize;
@@ -1711,12 +1712,13 @@ static pointObj point_norm(const pointObj a)
 {
   double lenmul;
   pointObj retv;
+  int norm_vector;
 
+  norm_vector = a.x==0 && a.y==0;
 #ifdef USE_POINT_Z_M
-  if (a.x==0 && a.y==0 && a.z==0 && a.m==0)
-#else
-  if (a.x==0 && a.y==0)
+  norm_vector = norm_vector && a.z==0 && a.m==0;
 #endif
+  if (norm_vector)
     return a;
 
   lenmul=1.0/sqrt(point_abs2(a));  /* this seems to be the costly operation */
@@ -1894,14 +1896,13 @@ int msSetup()
   msThreadInit();
 #endif
 
+  /* Use PROJ_LIB env vars if set */
+  msProjLibInitFromEnv();
+
   /* Use MS_ERRORFILE and MS_DEBUGLEVEL env vars if set */
   if (msDebugInitFromEnv() != MS_SUCCESS)
     return MS_FAILURE;
 
-#ifdef USE_GD
-  msGDSetup();
-#endif
-
 #ifdef USE_GEOS
   msGEOSSetup();
 #endif
@@ -1912,6 +1913,9 @@ int msSetup()
 #endif
 #endif
 
+  msFontCacheSetup();
+
+
   return MS_SUCCESS;
 }
 
@@ -1922,7 +1926,7 @@ int msSetup()
 #include "maplibxml2.h"
 #endif
 #endif
-void msCleanup(int signal)
+void msCleanup()
 {
   msForceTmpFileBase( NULL );
   msConnPoolFinalCleanup();
@@ -1950,10 +1954,6 @@ void msCleanup(int signal)
   msHTTPCleanup();
 #endif
 
-#ifdef USE_GD
-  msGDCleanup(signal);
-#endif
-
 #ifdef USE_GEOS
   msGEOSCleanup();
 #endif
@@ -1968,6 +1968,8 @@ void msCleanup(int signal)
 #endif
 #endif
 
+  msFontCacheCleanup();
+
   msTimeCleanup();
 
   msIO_Cleanup();
@@ -2096,41 +2098,79 @@ void msAlphaBlendPM( unsigned char red_src, unsigned char green_src,
   /*      Cases with actual blending.                                     */
   /* -------------------------------------------------------------------- */
   if(!alpha_dst || *alpha_dst == 255) {
-    int weight_dst = 256 - alpha_src;
+    int weight_dst = 255 - alpha_src;
 
-    *red_dst   = (256 * red_src   + *red_dst   * weight_dst) >> 8;
-    *green_dst = (256 * green_src + *green_dst * weight_dst) >> 8;
-    *blue_dst  = (256 * blue_src  + *blue_dst  * weight_dst) >> 8;
+    *red_dst   = (alpha_src * red_src   + *red_dst   * weight_dst) >> 8;
+    *green_dst = (alpha_src * green_src + *green_dst * weight_dst) >> 8;
+    *blue_dst  = (alpha_src * blue_src  + *blue_dst  * weight_dst) >> 8;
   } else {
-    int   weight_dst = (256 - alpha_src);
+    int   weight_dst = (255 - alpha_src);
 
-    *red_dst   = (256 * red_src   + *red_dst   * weight_dst) >> 8;
-    *green_dst = (256 * green_src + *green_dst * weight_dst) >> 8;
-    *blue_dst  = (256 * blue_src  + *blue_dst  * weight_dst) >> 8;
+    *red_dst   = (alpha_src * red_src   + *red_dst   * weight_dst) >> 8;
+    *green_dst = (alpha_src * green_src + *green_dst * weight_dst) >> 8;
+    *blue_dst  = (alpha_src * blue_src  + *blue_dst  * weight_dst) >> 8;
 
-    *alpha_dst = (256 * alpha_src + *alpha_dst * weight_dst) >> 8;
+    *alpha_dst = (255 * alpha_src + *alpha_dst * weight_dst) >> 8;
   }
 }
 
+void msRGBtoHSL(colorObj *rgb, double *h, double *s, double *l) {
+  double r = rgb->red/255.0, g = rgb->green/255.0, b = rgb->blue/255.0;
+  double maxv = MS_MAX(MS_MAX(r, g), b), minv = MS_MIN(MS_MIN(r, g), b);
+  double d = maxv - minv;
+  
+  *h = 0, *s = 0;
+  *l = (maxv + minv) / 2;
+  
+  if (maxv != minv)
+  {
+    *s = *l > 0.5 ? d / (2 - maxv - minv) : d / (maxv + minv);
+    if (maxv == r) { *h = (g - b) / d + (g < b ? 6 : 0); }
+    else if (maxv == g) { *h = (b - r) / d + 2; }
+    else if (maxv == b) { *h = (r - g) / d + 4; }
+    *h /= 6;
+  }
+}
+
+static double hue_to_rgb(double p, double q, double t) {
+  if(t < 0) t += 1;
+  if(t > 1) t -= 1;
+  if(t < 0.1666666666666666) return p + (q - p) * 6 * t;
+  if(t < 0.5) return q;
+  if(t < 0.6666666666666666) return p + (q - p) * (0.666666666666 - t) * 6;
+  return p;
+}
+
+void msHSLtoRGB(double h, double s, double l, colorObj *rgb) {
+  double r, g, b;
+  
+  if(s == 0){
+    r = g = b = l;
+  } else {
+    
+    double q = l < 0.5 ? l * (1 + s) : l + s - l * s;
+    double p = 2 * l - q;
+    r = hue_to_rgb(p, q, h + 0.33333333333333333);
+    g = hue_to_rgb(p, q, h);
+    b = hue_to_rgb(p, q, h - 0.33333333333333333);
+  }
+  rgb->red = r * 255;
+  rgb->green = g * 255;
+  rgb->blue = b * 255;
+}
+
 /*
  RFC 24: check if the parent pointer is NULL and raise an error otherwise
 */
 int msCheckParentPointer(void* p, char *objname)
 {
-  char* fmt="The %s parent object is null";
   char* msg=NULL;
   if (p == NULL) {
     if(objname != NULL) {
-      msg=malloc( sizeof(char) * ( ( strlen(fmt)+strlen(objname) ) ) );
-      if(msg == NULL) {
-        msg="A required parent object is null";
-      } else {
-        sprintf(msg, "The %s parent object is null", objname);
-      }
+      msSetError(MS_NULLPARENTERR, "The %s parent object is null", "msCheckParentPointer()", msg);
     } else {
-      msg="A required parent object is null";
+      msSetError(MS_NULLPARENTERR, "The parent object is null", "msCheckParentPointer()");
     }
-    msSetError(MS_NULLPARENTERR, "%s", "", msg);
     return MS_FAILURE;
   }
   return MS_SUCCESS;
@@ -2182,12 +2222,6 @@ void msFreeRasterBuffer(rasterBufferObj *b)
       b->data.palette.pixels = NULL;
       b->data.palette.palette = NULL;
       break;
-#ifdef USE_GD
-    case MS_BUFFER_GD:
-      gdImageDestroy(b->data.gd_img);
-      b->data.gd_img = NULL;
-      break;
-#endif
   }
 
 }
@@ -2263,11 +2297,11 @@ void *msSmallMalloc( size_t nSize )
 {
   void        *pReturn;
 
-  if( nSize == 0 )
+  if( UNLIKELY(nSize == 0) )
     return NULL;
 
   pReturn = malloc( nSize );
-  if( pReturn == NULL ) {
+  if( UNLIKELY(pReturn == NULL) ) {
     msIO_fprintf(stderr, "msSmallMalloc(): Out of memory allocating %ld bytes.\n",
                  (long) nSize );
     exit(1);
@@ -2286,12 +2320,12 @@ void * msSmallRealloc( void * pData, size_t nNewSize )
 {
   void        *pReturn;
 
-  if ( nNewSize == 0 )
+  if ( UNLIKELY(nNewSize == 0) )
     return NULL;
 
   pReturn = realloc( pData, nNewSize );
 
-  if( pReturn == NULL ) {
+  if( UNLIKELY(pReturn == NULL) ) {
     msIO_fprintf(stderr, "msSmallRealloc(): Out of memory allocating %ld bytes.\n",
                  (long)nNewSize );
     exit(1);
@@ -2310,11 +2344,11 @@ void *msSmallCalloc( size_t nCount, size_t nSize )
 {
   void  *pReturn;
 
-  if( nSize * nCount == 0 )
+  if( UNLIKELY(nSize * nCount == 0) )
     return NULL;
 
   pReturn = calloc( nCount, nSize );
-  if( pReturn == NULL ) {
+  if( UNLIKELY(pReturn == NULL) ) {
     msIO_fprintf(stderr, "msSmallCalloc(): Out of memory allocating %ld bytes.\n",
                  (long)(nCount*nSize));
     exit(1);
@@ -2337,10 +2371,23 @@ char *msBuildOnlineResource(mapObj *map, cgiRequestObj *req)
 {
   char *online_resource = NULL;
   const char *value, *hostname, *port, *script, *protocol="http", *mapparam=NULL;
-  int mapparam_len = 0;
+  char **hostname_array = NULL;
+  int mapparam_len = 0, hostname_array_len = 0;
 
-  hostname = getenv("SERVER_NAME");
-  port = getenv("SERVER_PORT");
+  hostname = getenv("HTTP_X_FORWARDED_HOST");
+  if(!hostname)
+    hostname = getenv("SERVER_NAME");
+  else {
+    if(strchr(hostname,',')) {
+      hostname_array = msStringSplit(hostname,',', &hostname_array_len);
+      hostname = hostname_array[0];
+    }
+  }
+
+  port = getenv("HTTP_X_FORWARDED_PORT");
+  if(!port)
+    port = getenv("SERVER_PORT");
+  
   script = getenv("SCRIPT_NAME");
 
   /* HTTPS is set by Apache to "on" in an HTTPS server ... if not set */
@@ -2349,6 +2396,9 @@ char *msBuildOnlineResource(mapObj *map, cgiRequestObj *req)
        ((value=getenv("SERVER_PORT")) && atoi(value) == 443) ) {
     protocol = "https";
   }
+  if ( (value=getenv("HTTP_X_FORWARDED_PROTO")) ) {
+    protocol = value;
+  }
 
   /* If map=.. was explicitly set then we'll include it in onlineresource
    */
@@ -2382,6 +2432,9 @@ char *msBuildOnlineResource(mapObj *map, cgiRequestObj *req)
     msSetError(MS_CGIERR, "Impossible to establish server URL.", "msBuildOnlineResource()");
     return NULL;
   }
+  if(hostname_array) {
+    msFreeCharArray(hostname_array, hostname_array_len);
+  }
 
   return online_resource;
 }
@@ -2426,27 +2479,86 @@ int msMapSetLayerProjections(mapObj* map)
   }
 
   for(i=0; i<map->numlayers; i++) {
+    layerObj *lp = GET_LAYER(map,i);
     /* This layer is turned on and needs a projection? */
-    if (GET_LAYER(map, i)->projection.numargs <= 0 &&
-        GET_LAYER(map, i)->status != MS_OFF &&
-        GET_LAYER(map, i)->transform == MS_TRUE) {
+    if (lp->projection.numargs <= 0 &&
+        lp->status != MS_OFF &&
+        lp->transform == MS_TRUE) {
 
       /* Fetch main map projection string only now that we need it */
       if (mapProjStr == NULL)
         mapProjStr = msGetProjectionString(&(map->projection));
 
       /* Set the projection to the map file projection */
-      if (msLoadProjectionString(&(GET_LAYER(map, i)->projection), mapProjStr) != 0) {
-        msSetError(MS_CGIERR, "Unable to set projection on layer.", "msTileSetProjectionst()");
+      if (msLoadProjectionString(&(lp->projection), mapProjStr) != 0) {
+        msSetError(MS_CGIERR, "Unable to set projection on layer.", "msMapSetLayerProjections()");
         return(MS_FAILURE);
       }
-      GET_LAYER(map, i)->project = MS_TRUE;
+      lp->project = MS_TRUE;
+      if(lp->connection && IS_THIRDPARTY_LAYER_CONNECTIONTYPE(lp->connectiontype)) {
+        char **reflayers;
+        int numreflayers,j;
+        reflayers = msStringSplit(lp->connection,',',&numreflayers);
+        for(j=0; j<numreflayers; j++) {
+          int *lidx, nlidx;
+          /* first check layers referenced by group name */
+          lidx = msGetLayersIndexByGroup(map, reflayers[i], &nlidx);
+          if(lidx) {
+            int k;
+            for(k=0; k<nlidx; k++) {
+              layerObj *glp = GET_LAYER(map,lidx[k]);
+              if (glp->projection.numargs <= 0 && glp->transform == MS_TRUE) {
+
+                /* Set the projection to the map file projection */
+                if (msLoadProjectionString(&(glp->projection), mapProjStr) != 0) {
+                  msSetError(MS_CGIERR, "Unable to set projection on layer.", "msMapSetLayerProjections()");
+                  return(MS_FAILURE);
+                }
+                glp->project = MS_TRUE;
+              }
+            }
+            free(lidx);
+          } else {
+            /* group name did not match, check by layer name */
+            int layer_idx = msGetLayerIndex(map,lp->connection);
+            layerObj *glp = GET_LAYER(map,layer_idx);
+            if (glp->projection.numargs <= 0 && glp->transform == MS_TRUE) {
+
+              /* Set the projection to the map file projection */
+              if (msLoadProjectionString(&(glp->projection), mapProjStr) != 0) {
+                msSetError(MS_CGIERR, "Unable to set projection on layer.", "msMapSetLayerProjections()");
+                return(MS_FAILURE);
+              }
+              glp->project = MS_TRUE;
+            }
+          }
+        }
+        msFreeCharArray(reflayers, numreflayers);
+      }
     }
   }
   msFree(mapProjStr);
   return(MS_SUCCESS);
 }
 
+
+/************************************************************************
+ *                    msMapSetLanguageSpecificConnection                *
+ *                                                                      *
+ *   Override DATA and CONNECTION of each layer with their specific     *
+ *  variant for the specified language.                                 *
+ ************************************************************************/
+
+void msMapSetLanguageSpecificConnection(mapObj* map, const char* validated_language)
+{
+    int i;
+    for(i=0; i<map->numlayers; i++) {
+      layerObj *layer = GET_LAYER(map, i);
+      if(layer->data) layer->data = msCaseReplaceSubstring(layer->data, "%language%", validated_language);
+      if(layer->connection) layer->connection = msCaseReplaceSubstring(layer->connection, "%language%", validated_language);
+    }
+}
+
 /* Generalize a shape based of the tolerance.
    Ref: http://trac.osgeo.org/gdal/ticket/966 */
 shapeObj* msGeneralize(shapeObj *shape, double tolerance)
@@ -2504,3 +2616,11 @@ shapeObj* msGeneralize(shapeObj *shape, double tolerance)
    
   return newShape;
 }
+
+void msSetLayerOpacity(layerObj *layer, int opacity) {
+  if(!layer->compositer) {
+    layer->compositer = msSmallMalloc(sizeof(LayerCompositer));
+    initLayerCompositer(layer->compositer);
+  }
+  layer->compositer->opacity = opacity;
+}
