@@ -33,6 +33,7 @@
 #include "mapserver.h"
 #include "maperror.h"
 #include "mapthread.h"
+#include "mapows.h"
 #include "mapwcs.h"
 
 
@@ -57,14 +58,12 @@
 **
 */
 
-int msWCSException11(mapObj *map, const char *locator,
-                     const char *exceptionCode, const char *version)
+int msWCSException11(mapObj *map, const char *exceptionCode,
+                     const char *locator, const char *version)
 {
   int size = 0;
   char *errorString     = NULL;
-  char *errorMessage    = NULL;
   char *schemasLocation = NULL;
-  const char * encoding;
 
   xmlDocPtr  psDoc      = NULL;
   xmlNodePtr psRootNode = NULL;
@@ -73,32 +72,26 @@ int msWCSException11(mapObj *map, const char *locator,
 
   psNsOws = xmlNewNs(NULL, BAD_CAST "http://www.opengis.net/ows/1.1", BAD_CAST "ows");
 
-  encoding = msOWSLookupMetadata(&(map->web.metadata), "CO", "encoding");
   errorString = msGetErrorString("\n");
-  errorMessage = msEncodeHTMLEntities(errorString);
   schemasLocation = msEncodeHTMLEntities(msOWSGetSchemasLocation(map));
 
   psDoc = xmlNewDoc(BAD_CAST "1.0");
 
-  psRootNode = msOWSCommonExceptionReport(psNsOws, OWS_1_1_0, schemasLocation, version, msOWSGetLanguage(map, "exception"), exceptionCode, locator, errorMessage);
+  psRootNode = msOWSCommonExceptionReport(psNsOws, OWS_1_1_0, schemasLocation, version, msOWSGetLanguage(map, "exception"), exceptionCode, locator, errorString);
 
   xmlDocSetRootElement(psDoc, psRootNode);
 
   xmlNewNs(psRootNode, BAD_CAST "http://www.opengis.net/ows/1.1", BAD_CAST "ows");
 
-  if (encoding)
-    msIO_setHeader("Content-Type","text/xml; charset=%s", encoding);
-  else
-    msIO_setHeader("Content-Type","text/xml");
+  msIO_setHeader("Content-Type","text/xml; charset=UTF-8");
   msIO_sendHeaders();
 
-  xmlDocDumpFormatMemoryEnc(psDoc, &buffer, &size, (encoding ? encoding : "ISO-8859-1"), 1);
+  xmlDocDumpFormatMemoryEnc(psDoc, &buffer, &size, "UTF-8", 1);
 
   msIO_printf("%s", buffer);
 
   /*free buffer and the document */
   free(errorString);
-  free(errorMessage);
   free(schemasLocation);
   xmlFree(buffer);
   xmlFreeDoc(psDoc);
@@ -156,9 +149,6 @@ static char *msWCSGetFormatsList11( mapObj *map, layerObj *layer )
     for( i = 0; i < map->numoutputformats; i++ ) {
       switch( map->outputformatlist[i]->renderer ) {
           /* seeminly normal raster format */
-#ifdef USE_GD
-        case MS_RENDER_WITH_GD:
-#endif
         case MS_RENDER_WITH_AGG:
         case MS_RENDER_WITH_RAWDATA:
           tokens[numtokens++] = msStrdup(map->outputformatlist[i]->name);
@@ -376,7 +366,6 @@ int msWCSGetCapabilities11(mapObj *map, wcsParamsObj *params,
   xmlNodePtr psRootNode, psMainNode, psNode;
   char *identifier_list = NULL, *format_list = NULL;
   const char *updatesequence=NULL;
-  const char *encoding;
   xmlNsPtr psOwsNs, psXLinkNs;
   char *schemaLocation = NULL;
   char *xsi_schemaLocation = NULL;
@@ -393,17 +382,16 @@ int msWCSGetCapabilities11(mapObj *map, wcsParamsObj *params,
   /* -------------------------------------------------------------------- */
 
   updatesequence = msOWSLookupMetadata(&(map->web.metadata), "CO", "updatesequence");
-  encoding = msOWSLookupMetadata(&(map->web.metadata), "CO", "encoding");
 
   if (params->updatesequence != NULL) {
     i = msOWSNegotiateUpdateSequence(params->updatesequence, updatesequence);
     if (i == 0) { /* current */
       msSetError(MS_WCSERR, "UPDATESEQUENCE parameter (%s) is equal to server (%s)", "msWCSGetCapabilities11()", params->updatesequence, updatesequence);
-      return msWCSException11(map, "updatesequence", "CurrentUpdateSequence", params->version);
+      return msWCSException11(map, "CurrentUpdateSequence", "updatesequence", params->version);
     }
     if (i > 0) { /* invalid */
       msSetError(MS_WCSERR, "UPDATESEQUENCE parameter (%s) is higher than server (%s)", "msWCSGetCapabilities11()", params->updatesequence, updatesequence);
-      return msWCSException11(map, "updatesequence", "InvalidUpdateSequence", params->version);
+      return msWCSException11(map, "InvalidUpdateSequence", "updatesequence", params->version);
     }
   }
 
@@ -472,7 +460,7 @@ int msWCSGetCapabilities11(mapObj *map, wcsParamsObj *params,
       || strstr(params->section,"All") != NULL
       || strstr(params->section,"ServiceIdentification") != NULL ) {
     xmlAddChild(psRootNode, msOWSCommonServiceIdentification(
-                              psOwsNs, map, "OGC WCS", params->version, "CO"));
+                              psOwsNs, map, "OGC WCS", "2.0.1,1.1.1,1.0.0", "CO", NULL));
   }
 
   /*service provider*/
@@ -480,7 +468,7 @@ int msWCSGetCapabilities11(mapObj *map, wcsParamsObj *params,
       || strstr(params->section,"All") != NULL
       || strstr(params->section,"ServiceProvider") != NULL ) {
     xmlAddChild(psRootNode, msOWSCommonServiceProvider(
-                              psOwsNs, psXLinkNs, map, "CO"));
+                              psOwsNs, psXLinkNs, map, "CO", NULL));
   }
 
   /* -------------------------------------------------------------------- */
@@ -490,7 +478,7 @@ int msWCSGetCapabilities11(mapObj *map, wcsParamsObj *params,
   if ((script_url=msOWSGetOnlineResource(map, "CO", "onlineresource", req)) == NULL
       || (script_url_encoded = msEncodeHTMLEntities(script_url)) == NULL) {
     msSetError(MS_WCSERR, "Server URL not found", "msWCSGetCapabilities11()");
-    return msWCSException11(map, "mapserv", "NoApplicableCode", params->version);
+    return msWCSException11(map, "NoApplicableCode", "mapserv", params->version);
   }
   free( script_url );
 
@@ -587,7 +575,10 @@ int msWCSGetCapabilities11(mapObj *map, wcsParamsObj *params,
 
         status = msWCSGetCapabilities11_CoverageSummary(
                    map, params, req, psDoc, psMainNode, layer );
-        if(status != MS_SUCCESS) return MS_FAILURE;
+        if(status != MS_SUCCESS) {
+          msFree(identifier_list);
+          return MS_FAILURE;
+        }
       }
     }
   }
@@ -599,15 +590,12 @@ int msWCSGetCapabilities11(mapObj *map, wcsParamsObj *params,
   if( msIO_needBinaryStdout() == MS_FAILURE )
     return MS_FAILURE;
 
-  if (encoding)
-    msIO_setHeader("Content-Type","text/xml; charset=%s", encoding);
-  else
-    msIO_setHeader("Content-Type","text/xml");
+  msIO_setHeader("Content-Type","text/xml; charset=UTF-8");
   msIO_sendHeaders();
 
   context = msIO_getHandler(stdout);
 
-  xmlDocDumpFormatMemoryEnc(psDoc, &buffer, &size, (encoding ? encoding : "ISO-8859-1"), 1);
+  xmlDocDumpFormatMemoryEnc(psDoc, &buffer, &size, "UTF-8", 1);
   msIO_contextWrite(context, buffer, size);
   xmlFree(buffer);
 
@@ -898,12 +886,9 @@ int msWCSDescribeCoverage11(mapObj *map, wcsParamsObj *params, owsRequestObj *ow
   xmlNsPtr psOwsNs;
   char *schemaLocation = NULL;
   char *xsi_schemaLocation = NULL;
-  const char *encoding;
 
   int i,j;
 
-  encoding = msOWSLookupMetadata(&(map->web.metadata), "CO", "encoding");
-
   /* -------------------------------------------------------------------- */
   /*      We will actually get the coverages list as a single item in     */
   /*      a string list with that item having the comma delimited         */
@@ -928,7 +913,7 @@ int msWCSDescribeCoverage11(mapObj *map, wcsParamsObj *params, owsRequestObj *ow
         msSetError( MS_WCSERR,
                     "COVERAGE %s cannot be opened / does not exist",
                     "msWCSDescribeCoverage()", params->coverages[j]);
-        return msWCSException11(map, "coverage", "CoverageNotDefined", params->version);
+        return msWCSException11(map, "CoverageNotDefined", "coverage", params->version);
       }
     }
   }
@@ -998,15 +983,12 @@ int msWCSDescribeCoverage11(mapObj *map, wcsParamsObj *params, owsRequestObj *ow
     if( msIO_needBinaryStdout() == MS_FAILURE )
       return MS_FAILURE;
 
-    if (encoding)
-      msIO_setHeader("Content-Type","text/xml; charset=%s", encoding);
-    else
-      msIO_setHeader("Content-Type","text/xml");
+    msIO_setHeader("Content-Type","text/xml; charset=UTF-8");
     msIO_sendHeaders();
 
     context = msIO_getHandler(stdout);
 
-    xmlDocDumpFormatMemoryEnc(psDoc, &buffer, &size, (encoding ? encoding : "ISO-8859-1"), 1);
+    xmlDocDumpFormatMemoryEnc(psDoc, &buffer, &size, "UTF-8", 1);
     msIO_contextWrite(context, buffer, size);
     xmlFree(buffer);
   }
@@ -1077,8 +1059,11 @@ int msWCSGetCoverageBands11( mapObj *map, cgiRequestObj *request,
   /* -------------------------------------------------------------------- */
   value = rangesubset + strlen(field_id);
 
-  if( strcasecmp(rangesubset,field_id) == 0 )
+  if( strcasecmp(rangesubset,field_id) == 0 ) {
+    free(rangesubset);
+    free(field_id);
     return MS_SUCCESS; /* we only got field ... default options */
+  }
 
   if( strlen(rangesubset) <= strlen(field_id)+1
       || strncasecmp(rangesubset,field_id,strlen(field_id)) != 0
@@ -1087,7 +1072,9 @@ int msWCSGetCoverageBands11( mapObj *map, cgiRequestObj *request,
                 "RangeSubset field name malformed, expected '%s', got RangeSubset=%s",
                 "msWCSGetCoverageBands11()",
                 field_id, rangesubset );
-    return msWCSException11(map, "mapserv", "NoApplicableCode", params->version);
+    free(rangesubset);
+    free(field_id);
+    return msWCSException11(map, "NoApplicableCode", "mapserv", params->version);
   }
 
   free( field_id );
@@ -1112,8 +1099,10 @@ int msWCSGetCoverageBands11( mapObj *map, cgiRequestObj *request,
   /* -------------------------------------------------------------------- */
   /*      Parse out the axis name, and verify.                            */
   /* -------------------------------------------------------------------- */
-  if( *value != '[' )
+  if( *value != '[' ) {
+    free(rangesubset);
     return MS_SUCCESS;
+  }
 
   value++;
 
@@ -1124,7 +1113,8 @@ int msWCSGetCoverageBands11( mapObj *map, cgiRequestObj *request,
                 "RangeSubset axis name malformed, expected '%s', got RangeSubset=%s",
                 "msWCSGetCoverageBands11()",
                 axis_id, rangesubset );
-    return msWCSException11(map, "mapserv", "NoApplicableCode", params->version);
+    free(rangesubset);
+    return msWCSException11(map, "NoApplicableCode", "mapserv", params->version);
   }
 
   /* -------------------------------------------------------------------- */
@@ -1141,7 +1131,7 @@ int msWCSGetCoverageBands11( mapObj *map, cgiRequestObj *request,
       break;
     }
   }
-
+  free(rangesubset);
   return MS_SUCCESS;
 }
 #endif
@@ -1160,11 +1150,8 @@ int  msWCSReturnCoverage11( wcsParamsObj *params, mapObj *map,
   int status, i;
   char *filename = NULL;
   char *base_dir = NULL;
-  const char *encoding;
   const char *fo_filename;
 
-  encoding = msOWSLookupMetadata(&(map->web.metadata), "CO", "encoding");
-
   fo_filename = msGetOutputFormatOption( image->format, "FILENAME", NULL );
 
   /* -------------------------------------------------------------------- */
@@ -1183,7 +1170,7 @@ int  msWCSReturnCoverage11( wcsParamsObj *params, mapObj *map,
                   "Failed to find %s driver.",
                   "msWCSReturnCoverage11()",
                   image->format->driver+5 );
-      return msWCSException11(map, "mapserv", "NoApplicableCode",
+      return msWCSException11(map, "NoApplicableCode", "mapserv",
                               params->version);
     }
 
@@ -1208,7 +1195,7 @@ int  msWCSReturnCoverage11( wcsParamsObj *params, mapObj *map,
         msFree(filename);
         msSetError(MS_MISCERR, "msSaveImage() failed",
                    "msWCSReturnCoverage11()");
-        return msWCSException11(map, "mapserv", "NoApplicableCode",
+        return msWCSException11(map, "NoApplicableCode", "mapserv",
                                 params->version);
       }
     }
@@ -1218,40 +1205,20 @@ int  msWCSReturnCoverage11( wcsParamsObj *params, mapObj *map,
   /* -------------------------------------------------------------------- */
   /*      Output stock header.                                            */
   /* -------------------------------------------------------------------- */
-  if (encoding) {
-    msIO_setHeader("Content-Type","multipart/mixed; boundary=wcs");
-    msIO_sendHeaders();
-    msIO_fprintf(
-      stdout,
-      "\r\n--wcs\r\n"
-      "Content-Type: text/xml; charset=%s\r\n"
-      "Content-ID: wcs.xml\r\n\r\n"
-      "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
-      "<Coverages\n"
-      "     xmlns=\"http://www.opengis.net/wcs/1.1\"\n"
-      "     xmlns:ows=\"http://www.opengis.net/ows/1.1\"\n"
-      "     xmlns:xlink=\"http://www.w3.org/1999/xlink\"\n"
-      "     xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\n"
-      "     xsi:schemaLocation=\"http://www.opengis.net/ows/1.1 ../owsCoverages.xsd\">\n"
-      "  <Coverage>\n",
-      encoding);
-  } else {
-    msIO_setHeader("Content-Type","multipart/mixed; boundary=wcs");
-    msIO_sendHeaders();
-    msIO_fprintf(
-      stdout,
-      "\r\n--wcs\r\n"
-      "Content-Type: text/xml\r\n"
-      "Content-ID: wcs.xml\r\n\r\n"
-      "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
-      "<Coverages\n"
-      "     xmlns=\"http://www.opengis.net/wcs/1.1\"\n"
-      "     xmlns:ows=\"http://www.opengis.net/ows/1.1\"\n"
-      "     xmlns:xlink=\"http://www.w3.org/1999/xlink\"\n"
-      "     xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\n"
-      "     xsi:schemaLocation=\"http://www.opengis.net/ows/1.1 ../owsCoverages.xsd\">\n"
-      "  <Coverage>\n");
-  }
+  msIO_setHeader("Content-Type","multipart/mixed; boundary=wcs");
+  msIO_sendHeaders();
+  msIO_printf(
+    "\r\n--wcs\r\n"
+    "Content-Type: text/xml; charset=UTF-8\r\n"
+    "Content-ID: wcs.xml\r\n\r\n"
+    "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
+    "<Coverages\n"
+    "     xmlns=\"http://www.opengis.net/wcs/1.1\"\n"
+    "     xmlns:ows=\"http://www.opengis.net/ows/1.1\"\n"
+    "     xmlns:xlink=\"http://www.w3.org/1999/xlink\"\n"
+    "     xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\n"
+    "     xsi:schemaLocation=\"http://www.opengis.net/ows/1.1 ../owsCoverages.xsd\">\n"
+    "  <Coverage>\n");
 
   /* -------------------------------------------------------------------- */
   /*      If we weren't able to write data under /vsimem, then we just    */
@@ -1276,7 +1243,7 @@ int  msWCSReturnCoverage11( wcsParamsObj *params, mapObj *map,
     status = msSaveImage(map, image, NULL);
     if( status != MS_SUCCESS ) {
       msSetError( MS_MISCERR, "msSaveImage() failed", "msWCSReturnCoverage11()");
-      return msWCSException11(map, "mapserv", "NoApplicableCode", params->version);
+      return msWCSException11(map, "NoApplicableCode", "mapserv", params->version);
     }
 
     msIO_fprintf( stdout, "\r\n--wcs--\r\n" );
@@ -1394,8 +1361,8 @@ int msWCSDescribeCoverage11(mapObj *map, wcsParamsObj *params,
 {
   msSetError( MS_WCSERR,
               "WCS 1.1 request made, but mapserver requires libxml2 for WCS 1.1 services and this is not configured.",
-              "msWCSDescribeCoverage11()", "NoApplicableCode" );
-  return msWCSException11(map, "mapserv", "NoApplicableCode", params->version);
+              "msWCSDescribeCoverage11()" );
+  return msWCSException11(map, "NoApplicableCode", "mapserv", params->version);
 }
 
 /* ==================================================================== */
@@ -1405,15 +1372,15 @@ int msWCSGetCapabilities11(mapObj *map, wcsParamsObj *params,
 {
   msSetError( MS_WCSERR,
               "WCS 1.1 request made, but mapserver requires libxml2 for WCS 1.1 services and this is not configured.",
-              "msWCSGetCapabilities11()", "NoApplicableCode" );
+              "msWCSGetCapabilities11()" );
 
-  return msWCSException11(map, "mapserv", "NoApplicableCode", params->version);
+  return msWCSException11(map, "NoApplicableCode", "mapserv", params->version);
 }
 
-int msWCSException11(mapObj *map, const char *locator, const char *exceptionCode, const char *version)
+int msWCSException11(mapObj *map, const char *exceptionCode, const char *locator, const char *version)
 {
   /* fallback to reporting using 1.0 style exceptions. */
-  return msWCSException( map, locator, exceptionCode, "1.0.0" );
+  return msWCSException( map, exceptionCode, locator, "1.0.0" );
 }
 
 #endif /* defined(USE_WCS_SVR) && !defined(USE_LIBXML2) */
